# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Iris-specific extensions to matplotlib, mimicking the :mod:`matplotlib.pyplot`
interface.

See also: :ref:`matplotlib <matplotlib:users-guide-index>`.

"""

import collections
import datetime

import cartopy.crs as ccrs
import cartopy.mpl.geoaxes
from cartopy.geodesic import Geodesic
import cftime
import matplotlib.axes
import matplotlib.collections as mpl_collections
import matplotlib.dates as mpl_dates
import matplotlib.pyplot as plt
from matplotlib.offsetbox import AnchoredText
import matplotlib.ticker as mpl_ticker
import matplotlib.transforms as mpl_transforms
import numpy as np
import numpy.ma as ma

import iris.cube
import iris.analysis.cartography as cartography
import iris.coords
import iris.coord_systems
from iris.exceptions import IrisError

# Importing iris.palette to register the brewer palettes.
import iris.palette
from iris.util import _meshgrid

# Cynthia Brewer citation text.
BREWER_CITE = "Colours based on ColorBrewer.org"

PlotDefn = collections.namedtuple("PlotDefn", ("coords", "transpose"))


def _get_plot_defn_custom_coords_picked(cube, coords, mode, ndims=2):
    def names(coords):
        result = []
        for coord in coords:
            if isinstance(coord, int):
                result.append("dim={}".format(coord))
            else:
                result.append(coord.name())
        return ", ".join(result)

    def as_coord(coord):
        if isinstance(coord, int):
            # Pass through valid dimension indexes.
            if coord >= ndims:
                emsg = (
                    "The data dimension ({}) is out of range for "
                    "the dimensionality of the required plot ({})"
                )
                raise IndexError(emsg.format(coord, ndims))
        else:
            coord = cube.coord(coord)
        return coord

    coords = list(map(as_coord, coords))

    # Check that we were given the right number of coordinates/dimensions.
    if len(coords) != ndims:
        raise ValueError(
            "The list of coordinates given (%s) should have the"
            " same length (%s) as the dimensionality of the"
            " required plot (%s)" % (names(coords), len(coords), ndims)
        )

    # Check which dimensions are spanned by each coordinate.
    def get_span(coord):
        if isinstance(coord, int):
            span = set([coord])
        else:
            span = set(cube.coord_dims(coord))
        return span

    spans = list(map(get_span, coords))
    for span, coord in zip(spans, coords):
        if not span:
            msg = "The coordinate {!r} doesn't span a data dimension."
            raise ValueError(msg.format(coord.name()))
        if mode == iris.coords.BOUND_MODE and len(span) not in [1, 2]:
            raise ValueError(
                "The coordinate {!r} has {} dimensions."
                "Cell-based plotting is only supported for"
                "coordinates with one or two dimensions.".format(
                    coord.name(), len(span)
                )
            )

    # Check the combination of coordinates spans enough (ndims) data
    # dimensions.
    total_span = set().union(*spans)
    if len(total_span) != ndims:
        raise ValueError(
            "The given coordinates ({}) don't span the {} data"
            " dimensions.".format(names(coords), ndims)
        )

    # If we have 2-dimensional data, and one or more 1-dimensional
    # coordinates, check if we need to transpose.
    transpose = False
    if ndims == 2 and min(map(len, spans)) == 1:
        for i, span in enumerate(spans):
            if len(span) == 1:
                if list(span)[0] == i:
                    transpose = True
                    break

    # Note the use of `reversed` to convert from the X-then-Y
    # convention of the end-user API to the V-then-U convention used by
    # the plotting routines.
    plot_coords = list(reversed(coords))
    return PlotDefn(plot_coords, transpose)


def _valid_bound_dim_coord(coord):
    result = None
    if coord and coord.ndim == 1 and coord.nbounds:
        result = coord
    return result


def _get_plot_defn(cube, mode, ndims=2):
    """
    Return data and plot-axis coords given a cube & a mode of either
    POINT_MODE or BOUND_MODE.

    """
    if cube.ndim != ndims:
        msg = "Cube must be %s-dimensional. Got %s dimensions."
        raise ValueError(msg % (ndims, cube.ndim))

    # Start by taking the DimCoords from each dimension.
    coords = [None] * ndims
    for dim_coord in cube.dim_coords:
        dim = cube.coord_dims(dim_coord)[0]
        coords[dim] = dim_coord

    # When appropriate, restrict to 1D with bounds.
    if mode == iris.coords.BOUND_MODE:
        coords = list(map(_valid_bound_dim_coord, coords))

    def guess_axis(coord):
        axis = None
        if coord is not None:
            axis = iris.util.guess_coord_axis(coord)
        return axis

    # Allow DimCoords in aux_coords to fill in for missing dim_coords.
    for dim, coord in enumerate(coords):
        if coord is None:
            aux_coords = cube.coords(dimensions=dim)
            aux_coords = [
                coord
                for coord in aux_coords
                if isinstance(coord, iris.coords.DimCoord)
            ]
            if aux_coords:
                aux_coords.sort(key=lambda coord: coord.metadata)
                coords[dim] = aux_coords[0]

    # If plotting a 2 dimensional plot, check for 2d coordinates
    if ndims == 2:
        missing_dims = [
            dim for dim, coord in enumerate(coords) if coord is None
        ]
        if missing_dims:
            # Note that this only picks up coordinates that span the dims
            two_dim_coords = cube.coords(dimensions=missing_dims)
            two_dim_coords = [
                coord for coord in two_dim_coords if coord.ndim == 2
            ]
            if len(two_dim_coords) >= 2:
                two_dim_coords.sort(key=lambda coord: coord.metadata)
                coords = two_dim_coords[:2]

    if mode == iris.coords.POINT_MODE:
        # Allow multi-dimensional aux_coords to override the dim_coords
        # along the Z axis. This results in a preference for using the
        # derived altitude over model_level_number or level_height.
        # Limit to Z axis to avoid preferring latitude over grid_latitude etc.
        axes = list(map(guess_axis, coords))
        axis = "Z"
        if axis in axes:
            for coord in cube.coords(dim_coords=False):
                if (
                    max(coord.shape) > 1
                    and iris.util.guess_coord_axis(coord) == axis
                ):
                    coords[axes.index(axis)] = coord

    # Re-order the coordinates to achieve the preferred
    # horizontal/vertical associations. If we can't associate
    # an axis to order the coordinates, fall back to using the cube dimension
    # followed by the name of the coordinate.
    def sort_key(coord):
        order = {"X": 2, "T": 1, "Y": -1, "Z": -2}
        axis = guess_axis(coord)
        return (
            order.get(axis, 0),
            coords.index(coord),
            coord and coord.name(),
        )

    sorted_coords = sorted(coords, key=sort_key)

    transpose = sorted_coords != coords
    return PlotDefn(sorted_coords, transpose)


def _can_draw_map(coords):
    std_names = [
        c and c.standard_name
        for c in coords
        if isinstance(c, iris.coords.Coord)
    ]
    valid_std_names = [
        ["latitude", "longitude"],
        ["grid_latitude", "grid_longitude"],
        ["projection_y_coordinate", "projection_x_coordinate"],
    ]
    return std_names in valid_std_names


def _broadcast_2d(u, v):
    # Matplotlib needs the U and V coordinates to have the same
    # dimensionality (either both 1D, or both 2D). So we simply
    # broadcast both to 2D to be on the safe side.
    u = np.atleast_2d(u)
    v = np.atleast_2d(v.T).T
    u, v = np.broadcast_arrays(u, v)
    return u, v


def _string_coord_axis_tick_labels(string_axes, axes=None):
    """Apply tick labels for string coordinates."""

    ax = axes if axes else plt.gca()
    for axis, ticks in string_axes.items():
        # Define a tick formatter. This will assign a label to all ticks
        # located precisely on  an integer in range(len(ticks)) and assign
        # an empty string to any other ticks.
        def ticker_func(tick_location, _):
            tick_locations = range(len(ticks))
            labels = ticks
            label_dict = dict(zip(tick_locations, labels))
            label = label_dict.get(tick_location, "")
            return label

        formatter = mpl_ticker.FuncFormatter(ticker_func)
        locator = mpl_ticker.MaxNLocator(integer=True)
        this_axis = getattr(ax, axis)
        this_axis.set_major_formatter(formatter)
        this_axis.set_major_locator(locator)


def _invert_yaxis(v_coord, axes=None):
    """
    Inverts the y-axis of the current plot based on conditions:

        * If the y-axis is already inverted we don't want to re-invert it.
        * If v_coord is None then it will not have any attributes.
        * If neither of the above are true then invert y if v_coord has
          attribute 'positive' set to 'down'.

    Args:

        * v_coord - the coord to be plotted on the y-axis

    """
    axes = axes if axes else plt.gca()
    yaxis_is_inverted = axes.yaxis_inverted()
    if not yaxis_is_inverted and isinstance(v_coord, iris.coords.Coord):
        attr_pve = v_coord.attributes.get("positive")
        if attr_pve is not None and attr_pve.lower() == "down":
            axes.invert_yaxis()


def _check_bounds_contiguity_and_mask(coord, data, atol=None, rtol=None):
    """
    Checks that any discontiguities in the bounds of the given coordinate only
    occur where the data is masked.

    Where a discontinuity occurs the grid created for plotting will not be
    correct. This does not matter if the data is masked in that location as
    this is not plotted.

    If a discontiguity occurs where the data is *not* masked, an error is
    raised.

    Args:
        coord: (iris.coord.Coord)
            Coordinate the bounds of which will be checked for contiguity
        data: (array)
            Data of the the cube we are plotting
        atol:
            Absolute tolerance when checking the contiguity. Defaults to None.
            If an absolute tolerance is not set, 1D coords are not checked (so
            as to not introduce a breaking change without a major release) but
            2D coords are always checked, by calling
            :meth:`iris.coords.Coord._discontiguity_in_bounds` with its default
            tolerance.

    """
    kwargs = {}
    data_is_masked = hasattr(data, "mask")
    if data_is_masked:
        # When checking the location of the discontiguities, we check against
        # the opposite of the mask, which is True where data exists.
        mask_invert = np.logical_not(data.mask)

    if coord.ndim == 1:
        # 1D coords are only checked if an absolute tolerance is set, to avoid
        # introducing a breaking change.
        if atol:
            contiguous, diffs = coord._discontiguity_in_bounds(atol=atol)

            if not contiguous and data_is_masked:
                not_masked_at_discontiguity = np.any(
                    np.logical_and(mask_invert[:-1], diffs)
                )
        else:
            return

    elif coord.ndim == 2:
        if atol:
            kwargs["atol"] = atol
        if rtol:
            kwargs["rtol"] = rtol
        contiguous, diffs = coord._discontiguity_in_bounds(**kwargs)

        if not contiguous and data_is_masked:
            diffs_along_x, diffs_along_y = diffs

            # Check along both dimensions that any discontiguous
            # points are correctly masked.
            not_masked_at_discontiguity_along_x = np.any(
                np.logical_and(mask_invert[:, :-1], diffs_along_x)
            )

            not_masked_at_discontiguity_along_y = np.any(
                np.logical_and(mask_invert[:-1], diffs_along_y)
            )

            not_masked_at_discontiguity = (
                not_masked_at_discontiguity_along_x
                or not_masked_at_discontiguity_along_y
            )

    # If any discontiguity occurs where the data is not masked the grid will be
    # created incorrectly, so raise an error.
    if not contiguous:
        if not data_is_masked:
            raise ValueError(
                "The bounds of the {} coordinate are not "
                "contiguous. Not able to create a suitable grid"
                "to plot. You can use "
                "iris.util.find_discontiguities() to identify "
                "discontiguities in your x and y coordinate "
                "bounds arrays.".format(coord.name())
            )
        if not_masked_at_discontiguity:
            raise ValueError(
                "The bounds of the {} coordinate are not "
                "contiguous and data is not masked where the "
                "discontiguity occurs. Not able to create a "
                "suitable grid to plot. You can use "
                "iris.util.find_discontiguities() to identify "
                "discontiguities in your x and y coordinate "
                "bounds arrays, and then mask them with "
                "iris.util.mask_cube()"
                "".format(coord.name())
            )


def _draw_2d_from_bounds(draw_method_name, cube, *args, **kwargs):
    # NB. In the interests of clarity we use "u" and "v" to refer to the
    # horizontal and vertical axes on the matplotlib plot.
    mode = iris.coords.BOUND_MODE
    # Get & remove the coords entry from kwargs.
    coords = kwargs.pop("coords", None)
    if coords is not None:
        plot_defn = _get_plot_defn_custom_coords_picked(
            cube, coords, mode, ndims=2
        )
    else:
        plot_defn = _get_plot_defn(cube, mode, ndims=2)

    contig_tol = kwargs.pop("contiguity_tolerance", None)

    for coord in plot_defn.coords:
        if hasattr(coord, "has_bounds") and coord.has_bounds():
            _check_bounds_contiguity_and_mask(
                coord, data=cube.data, atol=contig_tol
            )

    if _can_draw_map(plot_defn.coords):
        result = _map_common(
            draw_method_name,
            None,
            iris.coords.BOUND_MODE,
            cube,
            plot_defn,
            *args,
            **kwargs,
        )
    else:
        # Obtain data array.
        data = cube.data
        if plot_defn.transpose:
            data = data.T

        # Obtain U and V coordinates
        v_coord, u_coord = plot_defn.coords

        # Track numpy arrays to use for the actual plotting.
        plot_arrays = []

        # Map axis name to associated values.
        string_axes = {}

        for coord, axis_name, data_dim in zip(
            [u_coord, v_coord], ["xaxis", "yaxis"], [1, 0]
        ):
            if coord is None:
                values = np.arange(data.shape[data_dim] + 1)
            elif isinstance(coord, int):
                dim = 1 - coord if plot_defn.transpose else coord
                values = np.arange(data.shape[dim] + 1)
            else:
                if coord.points.dtype.char in "SU":
                    if coord.points.ndim != 1:
                        msg = "Coord {!r} must be one-dimensional."
                        raise ValueError(msg.format(coord))
                    if coord.bounds is not None:
                        msg = "Cannot plot bounded string coordinate."
                        raise ValueError(msg)
                    string_axes[axis_name] = coord.points
                    values = np.arange(data.shape[data_dim] + 1) - 0.5
                else:
                    values = coord.contiguous_bounds()
                    values = _fixup_dates(coord, values)
                    if values.dtype == np.dtype(object) and isinstance(
                        values[0], datetime.datetime
                    ):
                        values = mpl_dates.date2num(values)

            plot_arrays.append(values)

        u, v = plot_arrays

        # If the data is transposed, 2D coordinates will also need to be
        # transposed.
        if plot_defn.transpose is True:
            u, v = [coord.T if coord.ndim == 2 else coord for coord in [u, v]]

        if u.ndim == v.ndim == 1:
            u, v = _broadcast_2d(u, v)

        axes = kwargs.pop("axes", None)
        draw_method = getattr(axes if axes else plt, draw_method_name)
        result = draw_method(u, v, data, *args, **kwargs)

        # Apply tick labels for string coordinates.
        _string_coord_axis_tick_labels(string_axes, axes)

        # Invert y-axis if necessary.
        _invert_yaxis(v_coord, axes)

    return result


def _draw_2d_from_points(draw_method_name, arg_func, cube, *args, **kwargs):
    # NB. In the interests of clarity we use "u" and "v" to refer to the
    # horizontal and vertical axes on the matplotlib plot.
    mode = iris.coords.POINT_MODE
    # Get & remove the coords entry from kwargs.
    coords = kwargs.pop("coords", None)
    if coords is not None:
        plot_defn = _get_plot_defn_custom_coords_picked(cube, coords, mode)
    else:
        plot_defn = _get_plot_defn(cube, mode, ndims=2)

    if _can_draw_map(plot_defn.coords):
        result = _map_common(
            draw_method_name,
            arg_func,
            iris.coords.POINT_MODE,
            cube,
            plot_defn,
            *args,
            **kwargs,
        )
    else:
        # Obtain data array.
        data = cube.data
        if plot_defn.transpose:
            data = data.T
            # Also transpose the scatter marker color array,
            # as now mpl 2.x does not do this for free.
            if draw_method_name == "scatter" and "c" in kwargs:
                c = kwargs["c"]
                if hasattr(c, "T") and cube.data.shape == c.shape:
                    kwargs["c"] = c.T

        # Obtain U and V coordinates
        v_coord, u_coord = plot_defn.coords
        if u_coord is None:
            u = np.arange(data.shape[1])
        elif isinstance(u_coord, int):
            dim = 1 - u_coord if plot_defn.transpose else u_coord
            u = np.arange(data.shape[dim])
        else:
            u = u_coord.points
            u = _fixup_dates(u_coord, u)

        if v_coord is None:
            v = np.arange(data.shape[0])
        elif isinstance(v_coord, int):
            dim = 1 - v_coord if plot_defn.transpose else v_coord
            v = np.arange(data.shape[dim])
        else:
            v = v_coord.points
            v = _fixup_dates(v_coord, v)

        if plot_defn.transpose:
            u = u.T
            v = v.T

        # Track numpy arrays to use for the actual plotting.
        plot_arrays = []

        # Map axis name to associated values.
        string_axes = {}

        for values, axis_name in zip([u, v], ["xaxis", "yaxis"]):
            # Replace any string coordinates with "index" coordinates.
            if values.dtype.char in "SU":
                if values.ndim != 1:
                    raise ValueError(
                        "Multi-dimensional string coordinates "
                        "not supported."
                    )
                plot_arrays.append(np.arange(values.size))
                string_axes[axis_name] = values
            elif values.dtype == np.dtype(object) and isinstance(
                values[0], datetime.datetime
            ):
                plot_arrays.append(mpl_dates.date2num(values))
            else:
                plot_arrays.append(values)

        u, v = plot_arrays
        u, v = _broadcast_2d(u, v)

        axes = kwargs.pop("axes", None)
        draw_method = getattr(axes if axes else plt, draw_method_name)
        if arg_func is not None:
            args, kwargs = arg_func(u, v, data, *args, **kwargs)
            result = draw_method(*args, **kwargs)
        else:
            result = draw_method(u, v, data, *args, **kwargs)

        # Apply tick labels for string coordinates.
        _string_coord_axis_tick_labels(string_axes, axes)

        # Invert y-axis if necessary.
        _invert_yaxis(v_coord, axes)

    return result


def _fixup_dates(coord, values):
    if coord.units.calendar is not None and values.ndim == 1:
        # Convert coordinate values into tuples of
        # (year, month, day, hour, min, sec)
        dates = [coord.units.num2date(val).timetuple()[0:6] for val in values]
        if coord.units.calendar == "gregorian":
            r = [datetime.datetime(*date) for date in dates]
        else:
            try:
                import nc_time_axis
            except ImportError:
                msg = (
                    "Cannot plot against time in a non-gregorian "
                    'calendar, because "nc_time_axis" is not available :  '
                    "Install the package from "
                    "https://github.com/SciTools/nc-time-axis to enable "
                    "this usage."
                )
                raise IrisError(msg)

            r = [
                nc_time_axis.CalendarDateTime(
                    cftime.datetime(*date), coord.units.calendar
                )
                for date in dates
            ]
        values = np.empty(len(r), dtype=object)
        values[:] = r
    return values


def _data_from_coord_or_cube(c):
    if isinstance(c, iris.cube.Cube):
        data = c.data
    elif isinstance(c, iris.coords.Coord):
        data = _fixup_dates(c, c.points)
    else:
        raise TypeError("Plot arguments must be cubes or coordinates.")
    return data


def _uv_from_u_object_v_object(u_object, v_object):
    ndim_msg = "Cube or coordinate must be 1-dimensional. Got {} dimensions."
    if u_object is not None and u_object.ndim > 1:
        raise ValueError(ndim_msg.format(u_object.ndim))
    if v_object.ndim > 1:
        raise ValueError(ndim_msg.format(v_object.ndim))
    v = _data_from_coord_or_cube(v_object)
    if u_object is None:
        u = np.arange(v.shape[0])
    else:
        u = _data_from_coord_or_cube(u_object)
    return u, v


def _u_object_from_v_object(v_object):
    u_object = None
    if isinstance(v_object, iris.cube.Cube):
        plot_defn = _get_plot_defn(v_object, iris.coords.POINT_MODE, ndims=1)
        (u_object,) = plot_defn.coords
    return u_object


def _get_plot_objects(args):
    if len(args) > 1 and isinstance(
        args[1], (iris.cube.Cube, iris.coords.Coord)
    ):
        # two arguments
        u_object, v_object = args[:2]
        u, v = _uv_from_u_object_v_object(u_object, v_object)
        args = args[2:]
        if len(u) != len(v):
            msg = (
                "The x and y-axis objects are not compatible. They should "
                "have equal sizes but got ({}: {}) and ({}: {})."
            )
            raise ValueError(
                msg.format(u_object.name(), len(u), v_object.name(), len(v))
            )
    else:
        # single argument
        v_object = args[0]
        u_object = _u_object_from_v_object(v_object)
        u, v = _uv_from_u_object_v_object(u_object, args[0])
        args = args[1:]
    return u_object, v_object, u, v, args


def _get_geodesic_params(globe):
    # Derive the semimajor axis and flattening values for a given globe from
    # its attributes. If the values are under specified, raise a ValueError
    flattening = globe.flattening
    semimajor = globe.semimajor_axis
    try:
        if semimajor is None:
            # Has semiminor or raises error
            if flattening is None:
                # Has inverse flattening or raises error
                flattening = 1.0 / globe.inverse_flattening
            semimajor = globe.semiminor_axis / (1.0 - flattening)
        elif flattening is None:
            if globe.semiminor_axis is not None:
                flattening = (semimajor - globe.semiminor_axis) / float(
                    semimajor
                )
            else:
                # Has inverse flattening or raises error
                flattening = 1.0 / globe.inverse_flattening
    except TypeError:
        # One of the required attributes was None
        raise ValueError("The globe was underspecified.")

    return semimajor, flattening


def _shift_plot_sections(u_object, u, v):
    """
    Shifts subsections of u by multiples of 360 degrees within ranges
    defined by the points where the line should cross over the 0/360 degree
    longitude boundary.

    e.g. [ 300, 100, 200, 300, 100, 300 ] => [ 300, 460, 560, 660, 820, 660 ]

    """
    # Convert coordinates to true lat-lon
    src_crs = (
        u_object.coord_system.as_cartopy_crs()
        if u_object.coord_system is not None
        else ccrs.Geodetic()
    )
    tgt_crs = ccrs.Geodetic(globe=src_crs.globe)
    tgt_proj = ccrs.PlateCarree(globe=src_crs.globe)

    points = tgt_crs.transform_points(src_crs, u, v)
    startpoints = points[:-1, :2]
    endpoints = points[1:, :2]
    proj_x, proj_y, _ = tgt_proj.transform_points(src_crs, u, v).T

    # Calculate the inverse geodesic for each pair of points in turn, and
    # convert the start point's azimuth into a vector in the source coordinate
    # system.
    try:
        radius, flattening = _get_geodesic_params(src_crs.globe)
        geodesic = Geodesic(radius, flattening)
    except ValueError:
        geodesic = Geodesic()
    dists, azms, _ = geodesic.inverse(startpoints, endpoints).T
    azms_lon = np.sin(np.deg2rad(azms))
    azms_lat = np.cos(np.deg2rad(azms))
    azms_u, _ = src_crs.transform_vectors(
        tgt_proj, proj_x[:-1], proj_y[:-1], azms_lon, azms_lat
    )

    # Use the grid longitude values and the geodesic azimuth to determine
    # the points where the line should cross the 0/360 degree boundary, and
    # in which direction
    lwraps = np.logical_and(u[1:] > u[:-1], azms_u < 0)
    rwraps = np.logical_and(u[1:] < u[:-1], azms_u > 0)
    shifts = np.where(rwraps, 1, 0) - np.where(lwraps, 1, 0)
    shift_vals = shifts.cumsum() * u_object.units.modulus
    new_u = np.empty_like(u)
    new_u[0] = u[0]
    new_u[1:] = u[1:] + shift_vals
    return new_u


def _draw_1d_from_points(draw_method_name, arg_func, *args, **kwargs):
    # NB. In the interests of clarity we use "u" to refer to the horizontal
    # axes on the matplotlib plot and "v" for the vertical axes.

    # retrieve the objects that are plotted on the horizontal and vertical
    # axes (cubes or coordinates) and their respective values, along with the
    # argument tuple with these objects removed
    u_object, v_object, u, v, args = _get_plot_objects(args)

    # Track numpy arrays to use for the actual plotting.
    plot_arrays = []

    # Map axis name to associated values.
    string_axes = {}

    for values, axis_name in zip([u, v], ["xaxis", "yaxis"]):
        # Replace any string coordinates with "index" coordinates.
        if values.dtype.char in "SU":
            if values.ndim != 1:
                msg = "Multi-dimensional string coordinates are not supported."
                raise ValueError(msg)
            plot_arrays.append(np.arange(values.size))
            string_axes[axis_name] = values
        else:
            plot_arrays.append(values)

    u, v = plot_arrays

    # if both u_object and v_object are coordinates then check if a map
    # should be drawn
    if (
        isinstance(u_object, iris.coords.Coord)
        and isinstance(v_object, iris.coords.Coord)
        and _can_draw_map([v_object, u_object])
    ):
        # Replace non-cartopy subplot/axes with a cartopy alternative and set
        # the transform keyword.
        kwargs = _ensure_cartopy_axes_and_determine_kwargs(
            u_object, v_object, kwargs
        )
        if draw_method_name == "plot" and u_object.standard_name not in (
            "projection_x_coordinate",
            "projection_y_coordinate",
        ):
            u = _shift_plot_sections(u_object, u, v)

    axes = kwargs.pop("axes", None)
    draw_method = getattr(axes if axes else plt, draw_method_name)
    if arg_func is not None:
        args, kwargs = arg_func(u, v, *args, **kwargs)
        result = draw_method(*args, **kwargs)
    else:
        result = draw_method(u, v, *args, **kwargs)

    # Apply tick labels for string coordinates.
    _string_coord_axis_tick_labels(string_axes, axes)

    # Invert y-axis if necessary.
    _invert_yaxis(v_object, axes)

    return result


def _replace_axes_with_cartopy_axes(cartopy_proj):
    """
    Replace non-cartopy subplot/axes with a cartopy alternative
    based on the provided projection. If the current axes are already an
    instance of :class:`cartopy.mpl.geoaxes.GeoAxes` then no action is taken.

    """

    ax = plt.gca()
    if not isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
        fig = plt.gcf()
        if isinstance(ax, matplotlib.axes.SubplotBase):
            _ = fig.add_subplot(
                ax.get_subplotspec(),
                projection=cartopy_proj,
                title=ax.get_title(),
                xlabel=ax.get_xlabel(),
                ylabel=ax.get_ylabel(),
            )
        else:
            _ = fig.add_axes(
                projection=cartopy_proj,
                title=ax.get_title(),
                xlabel=ax.get_xlabel(),
                ylabel=ax.get_ylabel(),
            )

        # delete the axes which didn't have a cartopy projection
        fig.delaxes(ax)


def _ensure_cartopy_axes_and_determine_kwargs(x_coord, y_coord, kwargs):
    """
    Replace the current non-cartopy axes with :class:`cartopy.mpl.GeoAxes`
    and return the appropriate kwargs dict based on the provided coordinates
    and kwargs.

    """
    # Determine projection.
    if x_coord.coord_system != y_coord.coord_system:
        raise ValueError(
            "The X and Y coordinates must have equal coordinate" " systems."
        )
    cs = x_coord.coord_system
    if cs is not None:
        cartopy_proj = cs.as_cartopy_projection()
    else:
        cartopy_proj = ccrs.PlateCarree()

    # Ensure the current axes are a cartopy.mpl.GeoAxes instance.
    axes = kwargs.get("axes")
    if axes is None:
        if (
            isinstance(cs, iris.coord_systems.RotatedGeogCS)
            and x_coord.points.max() > 180
            and x_coord.points.max() < 360
            and x_coord.points.min() > 0
        ):
            # The RotatedGeogCS has 0 - 360 extent, different from the
            # assumptions made by Cartopy: rebase longitudes for the map axes
            # to set the datum longitude to the International Date Line.
            cs_kwargs = cs._ccrs_kwargs()
            cs_kwargs["central_rotated_longitude"] = 180.0
            adapted_cartopy_proj = ccrs.RotatedPole(**cs_kwargs)
            _replace_axes_with_cartopy_axes(adapted_cartopy_proj)
        else:
            _replace_axes_with_cartopy_axes(cartopy_proj)
    elif axes and not isinstance(axes, cartopy.mpl.geoaxes.GeoAxes):
        raise TypeError(
            "The supplied axes instance must be a cartopy " "GeoAxes instance."
        )

    # Set the "from transform" keyword.
    if "transform" in kwargs:
        raise ValueError(
            "The 'transform' keyword is not allowed as it "
            "automatically determined from the coordinate "
            "metadata."
        )
    new_kwargs = kwargs.copy()
    new_kwargs["transform"] = cartopy_proj

    return new_kwargs


def _check_geostationary_coords_and_convert(x, y, kwargs):
    # Geostationary stores projected coordinates as scanning angles (
    # radians), in line with CF definition (this behaviour is unique to
    # Geostationary). Before plotting, must be converted by multiplying by
    # satellite height.
    x, y = (i.copy() for i in (x, y))
    transform = kwargs.get("transform")
    if isinstance(transform, cartopy.crs.Geostationary):
        satellite_height = transform.proj4_params["h"]
        for i in (x, y):
            i *= satellite_height

    return x, y


def _map_common(
    draw_method_name, arg_func, mode, cube, plot_defn, *args, **kwargs
):
    """
    Draw the given cube on a map using its points or bounds.

    "Mode" parameter will switch functionality between POINT or BOUND plotting.


    """
    # Generate 2d x and 2d y grids.
    y_coord, x_coord = plot_defn.coords
    if mode == iris.coords.POINT_MODE:
        if x_coord.ndim == y_coord.ndim == 1:
            x, y = _meshgrid(x_coord.points, y_coord.points)
        elif x_coord.ndim == y_coord.ndim == 2:
            x = x_coord.points
            y = y_coord.points
        else:
            raise ValueError("Expected 1D or 2D XY coords")
    else:
        if not x_coord.ndim == y_coord.ndim == 2:
            try:
                x, y = _meshgrid(
                    x_coord.contiguous_bounds(), y_coord.contiguous_bounds()
                )
            # Exception translation.
            except iris.exceptions.CoordinateMultiDimError:
                raise ValueError(
                    "Expected two 1D coords. Could not get XY"
                    " grid from bounds. X or Y coordinate not"
                    " 1D."
                )
            except ValueError:
                raise ValueError(
                    "Could not get XY grid from bounds. "
                    "X or Y coordinate doesn't have 2 bounds "
                    "per point."
                )
        else:
            x = x_coord.contiguous_bounds()
            y = y_coord.contiguous_bounds()

    # Obtain the data array.
    data = cube.data
    if plot_defn.transpose:
        data = data.T

    # If we are global, then append the first column of data the array to the
    # last (and add 360 degrees) NOTE: if it is found that this block of code
    # is useful in anywhere other than this plotting routine, it may be better
    # placed in the CS.
    if getattr(x_coord, "circular", False):
        _, direction = iris.util.monotonic(
            x_coord.points, return_direction=True
        )
        y = np.append(y, y[:, 0:1], axis=1)
        x = np.append(x, x[:, 0:1] + 360 * direction, axis=1)
        data = ma.concatenate([data, data[:, 0:1]], axis=1)
        if "_v_data" in kwargs:
            v_data = kwargs["_v_data"]
            v_data = ma.concatenate([v_data, v_data[:, 0:1]], axis=1)
            kwargs["_v_data"] = v_data

    # Replace non-cartopy subplot/axes with a cartopy alternative and set the
    # transform keyword.
    kwargs = _ensure_cartopy_axes_and_determine_kwargs(
        x_coord, y_coord, kwargs
    )

    # Make Geostationary coordinates plot-able.
    x, y = _check_geostationary_coords_and_convert(x, y, kwargs)

    if arg_func is not None:
        new_args, kwargs = arg_func(x, y, data, *args, **kwargs)
    else:
        new_args = (x, y, data) + args

    # Draw the contour lines/filled contours.
    axes = kwargs.pop("axes", None)
    plotfn = getattr(axes if axes else plt, draw_method_name)
    return plotfn(*new_args, **kwargs)


def contour(cube, *args, **kwargs):
    """
    Draws contour lines based on the given Cube.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.contour` for details of other valid
    keyword arguments.

    """
    result = _draw_2d_from_points("contour", None, cube, *args, **kwargs)
    return result


def contourf(cube, *args, **kwargs):
    """
    Draws filled contours based on the given Cube.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.contourf` for details of other valid
    keyword arguments.

    """
    coords = kwargs.get("coords")
    kwargs.setdefault("antialiased", True)
    result = _draw_2d_from_points("contourf", None, cube, *args, **kwargs)

    # Matplotlib produces visible seams between anti-aliased polygons.
    # But if the polygons are virtually opaque then we can cover the seams
    # by drawing anti-aliased lines *underneath* the polygon joins.

    # Figure out the alpha level for the contour plot
    if result.alpha is None:
        alpha = result.collections[0].get_facecolor()[0][3]
    else:
        alpha = result.alpha
    # If the contours are anti-aliased and mostly opaque then draw lines under
    # the seams.
    if result.antialiased and alpha > 0.95:
        levels = result.levels
        colors = [c[0] for c in result.tcolors]
        if result.extend == "neither":
            levels = levels[1:-1]
            colors = colors[:-1]
        elif result.extend == "min":
            levels = levels[:-1]
            colors = colors[:-1]
        elif result.extend == "max":
            levels = levels[1:]
            colors = colors[:-1]
        else:
            colors = colors[:-1]
        if len(levels) > 0:
            # Draw the lines just *below* the polygons to ensure we minimise
            # any boundary shift.
            zorder = result.collections[0].zorder - 0.1
            axes = kwargs.get("axes", None)

            # Workaround for cartopy#1780.  We do not want contour to shrink
            # extent.
            if axes is None:
                _axes = plt.gca()
            else:
                _axes = axes

            # Subsequent calls to dataLim.update_from_data_xy should not ignore
            # current extent.
            _axes.dataLim.ignore(False)

            contour(
                cube,
                levels=levels,
                colors=colors,
                antialiased=True,
                zorder=zorder,
                coords=coords,
                axes=axes,
            )
            # Restore the current "image" to 'result' rather than the mappable
            # resulting from the additional call to contour().
            if axes:
                axes._sci(result)
            else:
                plt.sci(result)

    return result


def default_projection(cube):
    """
    Return the primary map projection for the given cube.

    Using the returned projection, one can create a cartopy map with::

        import matplotlib.pyplot as plt
        ax = plt.ax(projection=default_projection(cube))

    """
    # XXX logic seems flawed, but it is what map_setup did...
    cs = cube.coord_system("CoordSystem")
    projection = cs.as_cartopy_projection() if cs else None
    return projection


def default_projection_extent(cube, mode=iris.coords.POINT_MODE):
    """
    Return the cube's extents ``(x0, x1, y0, y1)`` in its default projection.

    Keyword arguments:

     * mode - Either ``iris.coords.POINT_MODE`` or ``iris.coords.BOUND_MODE``.
              Triggers whether the extent should be representative of the cell
              points, or the limits of the cell's bounds.
              The default is iris.coords.POINT_MODE.

    """
    extents = cartography._xy_range(cube, mode)
    xlim = extents[0]
    ylim = extents[1]
    return tuple(xlim) + tuple(ylim)


def _fill_orography(cube, coords, mode, vert_plot, horiz_plot, style_args):
    # Find the orography coordinate.
    orography = cube.coord("surface_altitude")

    if coords is not None:
        plot_defn = _get_plot_defn_custom_coords_picked(
            cube, coords, mode, ndims=2
        )
    else:
        plot_defn = _get_plot_defn(cube, mode, ndims=2)
    v_coord, u_coord = plot_defn.coords

    # Find which plot coordinate corresponds to the derived altitude, so that
    # we can replace altitude with the surface altitude.
    if v_coord and v_coord.standard_name == "altitude":
        # v is altitude, so plot u and orography with orog in the y direction.
        result = vert_plot(u_coord, orography, style_args)
    elif u_coord and u_coord.standard_name == "altitude":
        # u is altitude, so plot v and orography with orog in the x direction.
        result = horiz_plot(v_coord, orography, style_args)
    else:
        raise ValueError(
            "Plot does not use hybrid height. One of the "
            "coordinates to plot must be altitude, but %s and %s "
            "were given." % (u_coord.name(), v_coord.name())
        )
    return result


def orography_at_bounds(cube, facecolor="#888888", coords=None, axes=None):
    """Plots orography defined at cell boundaries from the given Cube."""

    # XXX Needs contiguous orography corners to work.
    raise NotImplementedError(
        "This operation is temporarily not provided "
        "until coordinates can expose 2d contiguous "
        "bounds (corners)."
    )

    style_args = {"edgecolor": "none", "facecolor": facecolor}

    def vert_plot(u_coord, orography, style_args):
        u = u_coord.contiguous_bounds()
        left = u[:-1]
        height = orography.points
        width = u[1:] - left
        plotfn = axes.bar if axes else plt.bar
        return plotfn(left, height, width, **style_args)

    def horiz_plot(v_coord, orography, style_args):
        v = v_coord.contiguous_bounds()
        bottom = v[:-1]
        width = orography.points
        height = v[1:] - bottom
        plotfn = axes.barh if axes else plt.barh
        return plotfn(bottom, width, height, **style_args)

    return _fill_orography(
        cube, coords, iris.coords.BOUND_MODE, vert_plot, horiz_plot, style_args
    )


def orography_at_points(cube, facecolor="#888888", coords=None, axes=None):
    """Plots orography defined at sample points from the given Cube."""

    style_args = {"facecolor": facecolor}

    def vert_plot(u_coord, orography, style_args):
        x = u_coord.points
        y = orography.points
        plotfn = axes.fill_between if axes else plt.fill_between
        return plotfn(x, y, **style_args)

    def horiz_plot(v_coord, orography, style_args):
        y = v_coord.points
        x = orography.points
        plotfn = axes.fill_betweenx if axes else plt.fill_betweenx
        return plotfn(y, x, **style_args)

    return _fill_orography(
        cube, coords, iris.coords.POINT_MODE, vert_plot, horiz_plot, style_args
    )


def outline(cube, coords=None, color="k", linewidth=None, axes=None):
    """
    Draws cell outlines based on the given Cube.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * color: None or mpl color The color of the cell outlines. If
        None, the matplotlibrc setting patch.edgecolor is used by
        default.

    * linewidth: None or number The width of the lines showing the
        cell outlines. If None, the default width in patch.linewidth
        in matplotlibrc is used.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    """
    result = _draw_2d_from_bounds(
        "pcolormesh",
        cube,
        facecolors="none",
        edgecolors=color,
        linewidth=linewidth,
        antialiased=True,
        coords=coords,
        axes=axes,
    )

    # set the _is_stroked property to get a single color grid.
    # See https://github.com/matplotlib/matplotlib/issues/1302
    result._is_stroked = False
    if hasattr(result, "_wrapped_collection_fix"):
        result._wrapped_collection_fix._is_stroked = False
    return result


def pcolor(cube, *args, **kwargs):
    """
    Draws a pseudocolor plot based on the given 2-dimensional Cube.

    The cube must have either two 1-dimensional coordinates or two
    2-dimensional coordinates with contiguous bounds to plot the cube against.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    * contiguity_tolerance: The absolute tolerance used when checking for
        contiguity between the bounds of the cells. Defaults to None.

    See :func:`matplotlib.pyplot.pcolor` for details of other valid
    keyword arguments.

    """
    kwargs.setdefault("antialiased", True)
    kwargs.setdefault("snap", False)
    result = _draw_2d_from_bounds("pcolor", cube, *args, **kwargs)
    return result


def pcolormesh(cube, *args, **kwargs):
    """
    Draws a pseudocolor plot based on the given 2-dimensional Cube.

    The cube must have either two 1-dimensional coordinates or two
    2-dimensional coordinates with contiguous bounds to plot against each
    other.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    * contiguity_tolerance: The absolute tolerance used when checking for
        contiguity between the bounds of the cells. Defaults to None.

    See :func:`matplotlib.pyplot.pcolormesh` for details of other
    valid keyword arguments.

    """
    result = _draw_2d_from_bounds("pcolormesh", cube, *args, **kwargs)
    return result


def points(cube, *args, **kwargs):
    """
    Draws sample point positions based on the given Cube.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or
        coordinate names. Use the given coordinates as the axes for the
        plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.scatter` for details of other valid
    keyword arguments.

    """

    def _scatter_args(u, v, data, *args, **kwargs):
        return ((u, v) + args, kwargs)

    return _draw_2d_from_points(
        "scatter", _scatter_args, cube, *args, **kwargs
    )


def _vector_component_args(x_points, y_points, u_data, *args, **kwargs):
    """
    Callback from _draw_2d_from_points for 'quiver' and 'streamlines'.

    Returns arguments (x, y, u, v), to be passed to the underlying matplotlib
    call.

    "u_data" will always be "u_cube.data".
    The matching "v_cube.data" component is stored in kwargs['_v_data'].

    """
    v_data = kwargs.pop("_v_data")

    # Rescale u+v values for plot distortion.
    crs = kwargs.get("transform", None)
    if crs:
        if not isinstance(crs, (ccrs.PlateCarree, ccrs.RotatedPole)):
            msg = (
                "Can only plot vectors provided in a lat-lon "
                'projection, i.e. equivalent to "cartopy.crs.PlateCarree" '
                'or "cartopy.crs.RotatedPole". This '
                "cube coordinate system translates as Cartopy {}."
            )
            raise ValueError(msg.format(crs))
        # Given the above check, the Y points must be latitudes.
        # We therefore **assume** they are in degrees : I'm not sure this
        # is wise, but all the rest of this plot code does that, e.g. in
        # _map_common.
        # TODO: investigate degree units assumptions, here + elsewhere.

        # Implement a latitude scaling, but preserve the given magnitudes.
        u_data, v_data = [arr.copy() for arr in (u_data, v_data)]
        mags = np.sqrt(u_data * u_data + v_data * v_data)
        v_data *= np.cos(np.deg2rad(y_points))
        scales = mags / np.sqrt(u_data * u_data + v_data * v_data)
        u_data *= scales
        v_data *= scales

    return ((x_points, y_points, u_data, v_data), kwargs)


def quiver(u_cube, v_cube, *args, **kwargs):
    """
    Draws an arrow plot from two vector component cubes.

    Args:

    * u_cube, v_cube : (:class:`~iris.cube.Cube`)
        u and v vector components.  Must have same shape and units.
        If the cubes have geographic coordinates, the values are treated as
        true distance differentials, e.g. windspeeds, and *not* map coordinate
        vectors.  The components are aligned with the North and East of the
        cube coordinate system.

    .. Note:

        At present, if u_cube and v_cube have geographic coordinates, then they
        must be in a lat-lon coordinate system, though it may be a rotated one.
        To transform wind values between coordinate systems, use
        :func:`iris.analysis.cartography.rotate_vectors`.
        To transform coordinate grid points, you will need to create
        2-dimensional arrays of x and y values.  These can be transformed with
        :meth:`cartopy.crs.CRS.transform_points`.

    Kwargs:

    * coords: (list of :class:`~iris.coords.Coord` or string)
        Coordinates or coordinate names. Use the given coordinates as the axes
        for the plot. The order of the given coordinates indicates which axis
        to use for each, where the first element is the horizontal
        axis of the plot and the second element is the vertical axis
        of the plot.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.quiver` for details of other valid
    keyword arguments.

    """
    #
    # TODO: check u + v cubes for compatibility.
    #
    kwargs["_v_data"] = v_cube.data
    return _draw_2d_from_points(
        "quiver", _vector_component_args, u_cube, *args, **kwargs
    )


def plot(*args, **kwargs):
    """
    Draws a line plot based on the given cube(s) or coordinate(s).

    The first one or two arguments may be cubes or coordinates to plot.
    Each of the following is valid::

        # plot a 1d cube against its dimension coordinate
        plot(cube)

        # plot a 1d coordinate
        plot(coord)

        # plot a 1d cube against a given 1d coordinate, with the cube
        # values on the y-axis and the coordinate on the x-axis
        plot(coord, cube)

        # plot a 1d cube against a given 1d coordinate, with the cube
        # values on the x-axis and the coordinate on the y-axis
        plot(cube, coord)

        # plot two 1d coordinates against one-another
        plot(coord1, coord2)

        # plot two 1d cubes against one-another
        plot(cube1, cube2)

    Kwargs:

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.plot` for details of additional valid
    keyword arguments.

    """
    if "coords" in kwargs:
        raise TypeError(
            '"coords" is not a valid plot keyword. Coordinates '
            "and cubes may be passed as arguments for "
            "full control of the plot axes."
        )
    _plot_args = None
    return _draw_1d_from_points("plot", _plot_args, *args, **kwargs)


def scatter(x, y, *args, **kwargs):
    """
    Draws a scatter plot based on the given cube(s) or coordinate(s).

    Args:

    * x: :class:`~iris.cube.Cube` or :class:`~iris.coords.Coord`
        A cube or a coordinate to plot on the x-axis.

    * y: :class:`~iris.cube.Cube` or :class:`~iris.coords.Coord`
        A cube or a coordinate to plot on the y-axis.

    Kwargs:

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    See :func:`matplotlib.pyplot.scatter` for details of additional
    valid keyword arguments.

    """
    # here we are more specific about argument types than generic 1d plotting
    if not isinstance(x, (iris.cube.Cube, iris.coords.Coord)):
        raise TypeError("x must be a cube or a coordinate.")
    if not isinstance(y, (iris.cube.Cube, iris.coords.Coord)):
        raise TypeError("y must be a cube or a coordinate.")
    args = (x, y) + args
    _plot_args = None
    return _draw_1d_from_points("scatter", _plot_args, *args, **kwargs)


# Provide convenience show method from pyplot
show = plt.show


def symbols(x, y, symbols, size, axes=None, units="inches"):
    """
    Draws fixed-size symbols.

    See :mod:`iris.symbols` for available symbols.

    Args:

    * x: iterable
        The x coordinates where the symbols will be plotted.

    * y: iterable
        The y coordinates where the symbols will be plotted.

    * symbols: iterable
        The symbols (from :mod:`iris.symbols`) to plot.

    * size: float
        The symbol size in `units`.

    Kwargs:

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    * units: ['inches', 'points']
        The unit for the symbol size.

    """
    if axes is None:
        axes = plt.gca()

    offsets = np.array(list(zip(x, y)))

    # XXX "match_original" doesn't work ... so brute-force it instead.
    #   PatchCollection constructor ignores all non-style keywords when using
    #   match_original
    #   See matplotlib.collections.PatchCollection.__init__
    #   Specifically matplotlib/collections line 1053
    # pc = PatchCollection(symbols, offsets=offsets, transOffset=ax.transData,
    #                      match_original=True)
    facecolors = [p.get_facecolor() for p in symbols]
    edgecolors = [p.get_edgecolor() for p in symbols]
    linewidths = [p.get_linewidth() for p in symbols]

    pc = mpl_collections.PatchCollection(
        symbols,
        offsets=offsets,
        transOffset=axes.transData,
        facecolors=facecolors,
        edgecolors=edgecolors,
        linewidths=linewidths,
    )

    if units == "inches":
        scale = axes.figure.dpi
    elif units == "points":
        scale = axes.figure.dpi / 72.0
    else:
        raise ValueError("Unrecognised units: '%s'" % units)
    pc.set_transform(mpl_transforms.Affine2D().scale(0.5 * size * scale))

    axes.add_collection(pc)
    axes.autoscale_view()


def citation(text, figure=None, axes=None):
    """
    Add a text citation to a plot.

    Places an anchored text citation in the bottom right
    hand corner of the plot.

    Args:

    * text:
        Citation text to be plotted.

    Kwargs:

    * figure:
        Target :class:`matplotlib.figure.Figure` instance. Defaults
        to the current figure if none provided.

    * axes: the :class:`matplotlib.axes.Axes` to use for drawing.
        Defaults to the current axes if none provided.

    """

    if text is not None and len(text):
        if figure is None and not axes:
            figure = plt.gcf()
        anchor = AnchoredText(text, prop=dict(size=6), frameon=True, loc=4)
        anchor.patch.set_boxstyle("round, pad=0, rounding_size=0.2")
        axes = axes if axes else figure.gca()
        axes.add_artist(anchor)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot.pcolormesh` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.tests.unit.plot import TestGraphicStringCoord
from iris.tests.unit.plot._blockplot_common import (
    MixinStringCoordPlot,
    Mixin2dCoordsPlot,
    Mixin2dCoordsContigTol,
)


if tests.MPL_AVAILABLE:
    import iris.plot as iplt

    PLOT_FUNCTION_TO_TEST = iplt.pcolormesh


@tests.skip_plot
class TestStringCoordPlot(MixinStringCoordPlot, TestGraphicStringCoord):
    def blockplot_func(self):
        return PLOT_FUNCTION_TO_TEST


@tests.skip_plot
class Test2dCoords(tests.IrisTest, Mixin2dCoordsPlot):
    def setUp(self):
        self.blockplot_setup()

    def blockplot_func(self):
        return PLOT_FUNCTION_TO_TEST


@tests.skip_plot
class Test2dContigTol(tests.IrisTest, Mixin2dCoordsContigTol):
    # Extra call kwargs expected -- unlike 'pcolor', there are none.
    additional_kwargs = {}

    def blockplot_func(self):
        return PLOT_FUNCTION_TO_TEST


if __name__ == "__main__":
    tests.main()



"""
Deriving Exner Pressure and Air Temperature
===========================================

This example shows some processing of cubes in order to derive further related
cubes; in this case the derived cubes are Exner pressure and air temperature
which are calculated by combining air pressure, air potential temperature and
specific humidity. Finally, the two new cubes are presented side-by-side in a
plot.

"""

import matplotlib.pyplot as plt
import matplotlib.ticker

import iris
import iris.coords as coords
import iris.iterate
import iris.plot as iplt
import iris.quickplot as qplt


def limit_colorbar_ticks(contour_object):
    """
    Takes a contour object which has an associated colorbar and limits the
    number of ticks on the colorbar to 4.

    """
    # Under Matplotlib v1.2.x the colorbar attribute of a contour object is
    # a tuple containing the colorbar and an axes object, whereas under
    # Matplotlib v1.3.x it is simply the colorbar.
    try:
        colorbar = contour_object.colorbar[0]
    except (AttributeError, TypeError):
        colorbar = contour_object.colorbar

    colorbar.locator = matplotlib.ticker.MaxNLocator(4)
    colorbar.update_ticks()


def main():
    fname = iris.sample_data_path("colpex.pp")

    # The list of phenomena of interest
    phenomena = ["air_potential_temperature", "air_pressure"]

    # Define the constraint on standard name and model level
    constraints = [
        iris.Constraint(phenom, model_level_number=1) for phenom in phenomena
    ]

    air_potential_temperature, air_pressure = iris.load_cubes(
        fname, constraints
    )

    # Define a coordinate which represents 1000 hPa
    p0 = coords.AuxCoord(1000, long_name="P0", units="hPa")
    # Convert reference pressure 'p0' into the same units as 'air_pressure'
    p0.convert_units(air_pressure.units)

    # Calculate Exner pressure
    exner_pressure = (air_pressure / p0) ** (287.05 / 1005.0)
    # Set the name (the unit is scalar)
    exner_pressure.rename("exner_pressure")

    # Calculate air_temp
    air_temperature = exner_pressure * air_potential_temperature
    # Set the name (the unit is K)
    air_temperature.rename("air_temperature")

    # Now create an iterator which will give us lat lon slices of
    # exner pressure and air temperature in the form
    # (exner_slice, air_temp_slice).
    lat_lon_slice_pairs = iris.iterate.izip(
        exner_pressure,
        air_temperature,
        coords=["grid_latitude", "grid_longitude"],
    )

    # For the purposes of this example, we only want to demonstrate the first
    # plot.
    lat_lon_slice_pairs = [next(lat_lon_slice_pairs)]

    plt.figure(figsize=(8, 4))
    for exner_slice, air_temp_slice in lat_lon_slice_pairs:
        plt.subplot(121)
        cont = qplt.contourf(exner_slice)

        # The default colorbar has a few too many ticks on it, causing text to
        # overlap. Therefore, limit the number of ticks.
        limit_colorbar_ticks(cont)

        plt.subplot(122)
        cont = qplt.contourf(air_temp_slice)
        limit_colorbar_ticks(cont)
        iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for loading and saving netcdf files."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from os.path import join as path_join, dirname, sep as os_sep
import shutil
from subprocess import check_call
import tempfile

import iris
from iris.tests import stock


class TestClimatology(iris.tests.IrisTest):
    reference_cdl_path = os_sep.join(
        [
            dirname(tests.__file__),
            (
                "results/integration/climatology/TestClimatology/"
                "reference_simpledata.cdl"
            ),
        ]
    )

    @classmethod
    def _simple_cdl_string(cls):
        with open(cls.reference_cdl_path, "r") as f:
            cdl_content = f.read()
        # Add the expected CDL first line since this is removed from the
        # stored results file.
        cdl_content = "netcdf {\n" + cdl_content

        return cdl_content

    @staticmethod
    def _load_sanitised_cube(filepath):
        cube = iris.load_cube(filepath)
        # Remove attributes convention, if any.
        cube.attributes.pop("Conventions", None)
        # Remove any var-names.
        for coord in cube.coords():
            coord.var_name = None
        cube.var_name = None
        return cube

    @classmethod
    def setUpClass(cls):
        # Create a temp directory for temp files.
        cls.temp_dir = tempfile.mkdtemp()
        cls.path_ref_cdl = path_join(cls.temp_dir, "standard.cdl")
        cls.path_ref_nc = path_join(cls.temp_dir, "standard.nc")
        # Create reference CDL file.
        with open(cls.path_ref_cdl, "w") as f_out:
            f_out.write(cls._simple_cdl_string())
        # Create reference netCDF file from reference CDL.
        command = "ncgen -o {} {}".format(cls.path_ref_nc, cls.path_ref_cdl)
        check_call(command, shell=True)
        cls.path_temp_nc = path_join(cls.temp_dir, "tmp.nc")

        # Create reference cube.
        cls.cube_ref = stock.climatology_3d()

    @classmethod
    def tearDownClass(cls):
        # Destroy a temp directory for temp files.
        shutil.rmtree(cls.temp_dir)

    ###############################################################################
    # Round-trip tests

    def test_cube_to_cube(self):
        # Save reference cube to file, load cube from same file, test against
        # reference cube.
        iris.save(self.cube_ref, self.path_temp_nc)
        cube = self._load_sanitised_cube(self.path_temp_nc)
        self.assertEqual(cube, self.cube_ref)

    def test_file_to_file(self):
        # Load cube from reference file, save same cube to file, test against
        # reference CDL.
        cube = iris.load_cube(self.path_ref_nc)
        iris.save(cube, self.path_temp_nc)
        self.assertCDL(
            self.path_temp_nc,
            reference_filename=self.reference_cdl_path,
            flags="",
        )

    # NOTE:
    # The saving half of the round-trip tests is tested in the
    # appropriate dedicated test class:
    # unit.fileformats.netcdf.test_Saver.Test_write.test_with_climatology .
    # The loading half has no equivalent dedicated location, so is tested
    # here as test_load_from_file.

    def test_load_from_file(self):
        # Create cube from file, test against reference cube.
        cube = self._load_sanitised_cube(self.path_ref_nc)
        self.assertEqual(cube, self.cube_ref)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris.experimental.regrid.regrid_weighted_curvilinear_to_rectilinear`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import copy

import numpy as np
import numpy.ma as ma

import iris
import iris.coords
from iris.coords import AuxCoord, DimCoord
from iris.coord_systems import GeogCS, LambertConformal
from iris.fileformats.pp import EARTH_RADIUS
import iris.cube
from iris.experimental.regrid import (
    regrid_weighted_curvilinear_to_rectilinear as regrid,
)


PLAIN_LATLON_CS = GeogCS(EARTH_RADIUS)


class Test(tests.IrisTest):
    def setUp(self):
        # Source cube.
        self.test_src_name = "air_temperature"
        self.test_src_units = "K"
        self.test_src_data = ma.arange(1, 13, dtype=np.float).reshape(3, 4)
        self.test_src_attributes = dict(wibble="wobble")
        self.test_scalar_coord = iris.coords.DimCoord(
            [1], long_name="test_scalar_coord"
        )
        self.src = iris.cube.Cube(
            self.test_src_data,
            standard_name=self.test_src_name,
            units=self.test_src_units,
            aux_coords_and_dims=[(self.test_scalar_coord, None)],
            attributes=self.test_src_attributes,
        )

        # Source cube x-coordinates.
        points = np.array(
            [[10, 20, 200, 220], [110, 120, 180, 185], [190, 203, 211, 220]]
        )
        self.src_x_positive = iris.coords.AuxCoord(
            points,
            standard_name="longitude",
            units="degrees",
            coord_system=PLAIN_LATLON_CS,
        )
        self.src_x_transpose = iris.coords.AuxCoord(
            points.T,
            standard_name="longitude",
            units="degrees",
            coord_system=PLAIN_LATLON_CS,
        )
        points = np.array(
            [
                [-180, -176, -170, -150],
                [-180, -179, -178, -177],
                [-170, -168, -159, -140],
            ]
        )
        self.src_x_negative = iris.coords.AuxCoord(
            points,
            standard_name="longitude",
            units="degrees",
            coord_system=PLAIN_LATLON_CS,
        )

        # Source cube y-coordinates.
        points = np.array([[0, 4, 3, 1], [5, 7, 10, 6], [12, 20, 15, 30]])
        self.src_y = iris.coords.AuxCoord(
            points,
            standard_name="latitude",
            units="degrees",
            coord_system=PLAIN_LATLON_CS,
        )
        self.src_y_transpose = iris.coords.AuxCoord(
            points.T,
            standard_name="latitude",
            units="degrees",
            coord_system=PLAIN_LATLON_CS,
        )

        # Weights.
        self.weight_factor = 10
        self.weights = np.asarray(self.test_src_data) * self.weight_factor

        # Target grid cube.
        self.grid = iris.cube.Cube(np.zeros((2, 2)))

        # Target grid cube x-coordinates.
        self.grid_x_inc = iris.coords.DimCoord(
            [187, 200],
            standard_name="longitude",
            units="degrees",
            bounds=[[180, 190], [190, 220]],
            coord_system=PLAIN_LATLON_CS,
        )
        self.grid_x_dec = iris.coords.DimCoord(
            [200, 187],
            standard_name="longitude",
            units="degrees",
            bounds=[[220, 190], [190, 180]],
            coord_system=PLAIN_LATLON_CS,
        )

        # Target grid cube y-coordinates.
        self.grid_y_inc = iris.coords.DimCoord(
            [2, 10],
            standard_name="latitude",
            units="degrees",
            bounds=[[0, 5], [5, 30]],
            coord_system=PLAIN_LATLON_CS,
        )
        self.grid_y_dec = iris.coords.DimCoord(
            [10, 2],
            standard_name="latitude",
            units="degrees",
            bounds=[[30, 5], [5, 0]],
            coord_system=PLAIN_LATLON_CS,
        )

    def _weighted_mean(self, points):
        points = np.asarray(points, dtype=np.float)
        weights = points * self.weight_factor
        numerator = denominator = 0
        for point, weight in zip(points, weights):
            numerator += point * weight
            denominator += weight
        return numerator / denominator

    def _expected_cube(self, data):
        cube = iris.cube.Cube(data)
        cube.metadata = copy.deepcopy(self.src.metadata)
        grid_x = self.grid.coord(axis="x")
        grid_y = self.grid.coord(axis="y")
        cube.add_dim_coord(grid_x.copy(), self.grid.coord_dims(grid_x))
        cube.add_dim_coord(grid_y.copy(), self.grid.coord_dims(grid_y))
        src_x = self.src.coord(axis="x")
        src_y = self.src.coord(axis="y")
        for coord in self.src.aux_coords:
            if coord is not src_x and coord is not src_y:
                if not self.src.coord_dims(coord):
                    cube.add_aux_coord(coord)
        return cube

    def test_aligned_src_x(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_positive, (0, 1))
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                0,
                self._weighted_mean([3]),
                self._weighted_mean([7, 8]),
                self._weighted_mean([9, 10, 11]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[True, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_non_latlon(self):
        odd_coord_system = LambertConformal()
        co_src_y = AuxCoord(
            self.src_y.points,
            standard_name="projection_y_coordinate",
            units="km",
            coord_system=odd_coord_system,
        )
        co_src_x = AuxCoord(
            self.src_x_positive.points,
            standard_name="projection_x_coordinate",
            units="km",
            coord_system=odd_coord_system,
        )
        co_grid_y = DimCoord(
            self.grid_y_inc.points,
            bounds=self.grid_y_inc.bounds,
            standard_name="projection_y_coordinate",
            units="km",
            coord_system=odd_coord_system,
        )
        co_grid_x = DimCoord(
            self.grid_x_inc.points,
            bounds=self.grid_x_inc.bounds,
            standard_name="projection_x_coordinate",
            units="km",
            coord_system=odd_coord_system,
        )
        self.src.add_aux_coord(co_src_y, (0, 1))
        self.src.add_aux_coord(co_src_x, (0, 1))
        self.grid.add_dim_coord(co_grid_y, 0)
        self.grid.add_dim_coord(co_grid_x, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                0,
                self._weighted_mean([3]),
                self._weighted_mean([7, 8]),
                self._weighted_mean([9, 10, 11]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[True, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_src_xy_not_2d(self):
        new_shape = (2, 2, 3)
        # Reshape the source cube, including the X and Y coordinates,
        # from (3, 4) to (2, 2, 3).
        # This is really an "invalid" reshape, but should still work because
        # the XY shape is actually irrelevant to the regrid operation.
        src = iris.cube.Cube(
            self.test_src_data.reshape(new_shape),
            standard_name=self.test_src_name,
            units=self.test_src_units,
            aux_coords_and_dims=[(self.test_scalar_coord, None)],
            attributes=self.test_src_attributes,
        )
        co_src_y = self.src_y.copy(points=self.src_y.points.reshape(new_shape))
        co_src_x = self.src_x_positive.copy(
            points=self.src_x_positive.points.reshape(new_shape)
        )
        src.add_aux_coord(co_src_y, (0, 1, 2))
        src.add_aux_coord(co_src_x, (0, 1, 2))
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        weights = self.weights.reshape(new_shape)
        result = regrid(src, weights, self.grid)
        # NOTE: set the grid of self.src to make '_expected_cube' work ...
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_positive, (0, 1))
        # ... given that, we expect exactly the same 'normal' result.
        data = np.array(
            [
                0,
                self._weighted_mean([3]),
                self._weighted_mean([7, 8]),
                self._weighted_mean([9, 10, 11]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[True, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_aligned_src_x_mask(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_positive, (0, 1))
        self.src.data[([1, 2, 2], [3, 0, 2])] = ma.masked
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                0,
                self._weighted_mean([3]),
                self._weighted_mean([7]),
                self._weighted_mean([10]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[True, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_aligned_src_x_zero_weights(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_positive, (0, 1))
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        self.weights[:, 2] = 0
        self.weights[1, :] = 0
        result = regrid(self.src, self.weights, self.grid)
        data = np.array([0, 0, 0, self._weighted_mean([9, 10])]).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[True, True], [True, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_aligned_tgt_dec(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_positive, (0, 1))
        self.grid.add_dim_coord(self.grid_y_dec, 0)
        self.grid.add_dim_coord(self.grid_x_dec, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                self._weighted_mean([10, 11, 12]),
                self._weighted_mean([8, 9]),
                self._weighted_mean([3, 4]),
                0,
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[False, False], [False, True]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_misaligned_src_x_negative(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_negative, (0, 1))
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                self._weighted_mean([1, 2]),
                self._weighted_mean([3, 4]),
                self._weighted_mean([5, 6, 7, 8]),
                self._weighted_mean([9, 10, 11]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[False, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_misaligned_src_x_negative_mask(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_negative, (0, 1))
        self.src.data[([0, 0, 1, 1, 2, 2], [1, 3, 1, 3, 1, 3])] = ma.masked
        self.grid.add_dim_coord(self.grid_y_inc, 0)
        self.grid.add_dim_coord(self.grid_x_inc, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                self._weighted_mean([1]),
                self._weighted_mean([3]),
                self._weighted_mean([5, 7]),
                self._weighted_mean([9, 11]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[False, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)

    def test_misaligned_tgt_dec(self):
        self.src.add_aux_coord(self.src_y, (0, 1))
        self.src.add_aux_coord(self.src_x_negative, (0, 1))
        self.grid.add_dim_coord(self.grid_y_dec, 0)
        self.grid.add_dim_coord(self.grid_x_dec, 1)
        result = regrid(self.src, self.weights, self.grid)
        data = np.array(
            [
                self._weighted_mean([10, 11, 12]),
                self._weighted_mean([6, 7, 8, 9]),
                self._weighted_mean([4]),
                self._weighted_mean([2, 3]),
            ]
        ).reshape(2, 2)
        expected = self._expected_cube(data)
        self.assertEqual(result, expected)
        mask = np.array([[False, False], [False, False]])
        self.assertArrayEqual(result.data.mask, mask)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.netcdf.Saver` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from contextlib import contextmanager
from unittest import mock

import netCDF4 as nc
import numpy as np
from numpy import ma

import iris
from iris._lazy_data import as_lazy_data
from iris.coord_systems import (
    GeogCS,
    TransverseMercator,
    RotatedGeogCS,
    LambertConformal,
    Mercator,
    Stereographic,
    LambertAzimuthalEqualArea,
    AlbersEqualArea,
    VerticalPerspective,
    Geostationary,
)
from iris.coords import DimCoord
from iris.cube import Cube
from iris.fileformats.netcdf import Saver
import iris.tests.stock as stock


class Test_write(tests.IrisTest):
    # -------------------------------------------------------------------------
    # It is not considered necessary to have integration tests for saving
    # EVERY coordinate system. A subset are tested below.
    # -------------------------------------------------------------------------
    def _transverse_mercator_cube(self, ellipsoid=None):
        data = np.arange(12).reshape(3, 4)
        cube = Cube(data, "air_pressure_anomaly")
        trans_merc = TransverseMercator(
            49.0, -2.0, -400000.0, 100000.0, 0.9996012717, ellipsoid
        )
        coord = DimCoord(
            np.arange(3),
            "projection_y_coordinate",
            units="m",
            coord_system=trans_merc,
        )
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(
            np.arange(4),
            "projection_x_coordinate",
            units="m",
            coord_system=trans_merc,
        )
        cube.add_dim_coord(coord, 1)
        return cube

    def _mercator_cube(self, ellipsoid=None):
        data = np.arange(12).reshape(3, 4)
        cube = Cube(data, "air_pressure_anomaly")
        merc = Mercator(49.0, ellipsoid)
        coord = DimCoord(
            np.arange(3),
            "projection_y_coordinate",
            units="m",
            coord_system=merc,
        )
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(
            np.arange(4),
            "projection_x_coordinate",
            units="m",
            coord_system=merc,
        )
        cube.add_dim_coord(coord, 1)
        return cube

    def _stereo_cube(self, ellipsoid=None):
        data = np.arange(12).reshape(3, 4)
        cube = Cube(data, "air_pressure_anomaly")
        stereo = Stereographic(
            -10.0, 20.0, 500000.0, -200000.0, None, ellipsoid
        )
        coord = DimCoord(
            np.arange(3),
            "projection_y_coordinate",
            units="m",
            coord_system=stereo,
        )
        cube.add_dim_coord(coord, 0)
        coord = DimCoord(
            np.arange(4),
            "projection_x_coordinate",
            units="m",
            coord_system=stereo,
        )
        cube.add_dim_coord(coord, 1)
        return cube

    def test_transverse_mercator(self):
        # Create a Cube with a transverse Mercator coordinate system.
        ellipsoid = GeogCS(6377563.396, 6356256.909)
        cube = self._transverse_mercator_cube(ellipsoid)
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def test_transverse_mercator_no_ellipsoid(self):
        # Create a Cube with a transverse Mercator coordinate system.
        cube = self._transverse_mercator_cube()
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def test_mercator(self):
        # Create a Cube with a Mercator coordinate system.
        ellipsoid = GeogCS(6377563.396, 6356256.909)
        cube = self._mercator_cube(ellipsoid)
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def test_stereographic(self):
        # Create a Cube with a stereographic coordinate system.
        ellipsoid = GeogCS(6377563.396, 6356256.909)
        cube = self._stereo_cube(ellipsoid)
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def test_mercator_no_ellipsoid(self):
        # Create a Cube with a Mercator coordinate system.
        cube = self._mercator_cube()
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def test_stereographic_no_ellipsoid(self):
        # Create a Cube with a stereographic coordinate system.
        cube = self._stereo_cube()
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)

    def _simple_cube(self, dtype):
        data = np.arange(12, dtype=dtype).reshape(3, 4)
        points = np.arange(3, dtype=dtype)
        bounds = np.arange(6, dtype=dtype).reshape(3, 2)
        cube = Cube(data, "air_pressure_anomaly")
        coord = DimCoord(points, bounds=bounds, units="1")
        cube.add_dim_coord(coord, 0)
        return cube

    def test_little_endian(self):
        # Create a Cube with little-endian data.
        cube = self._simple_cube("<f4")
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            result_path = self.result_path("endian", "cdl")
            self.assertCDL(nc_path, result_path, flags="")

    def test_big_endian(self):
        # Create a Cube with big-endian data.
        cube = self._simple_cube(">f4")
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            result_path = self.result_path("endian", "cdl")
            self.assertCDL(nc_path, result_path, flags="")

    def test_zlib(self):
        cube = self._simple_cube(">f4")
        api = self.patch("iris.fileformats.netcdf.netCDF4")
        with Saver("/dummy/path", "NETCDF4") as saver:
            saver.write(cube, zlib=True)
        dataset = api.Dataset.return_value
        create_var_call = mock.call(
            "air_pressure_anomaly",
            np.dtype("float32"),
            ["dim0", "dim1"],
            fill_value=None,
            shuffle=True,
            least_significant_digit=None,
            contiguous=False,
            zlib=True,
            fletcher32=False,
            endian="native",
            complevel=4,
            chunksizes=None,
        )
        self.assertIn(create_var_call, dataset.createVariable.call_args_list)

    def test_least_significant_digit(self):
        cube = Cube(
            np.array([1.23, 4.56, 7.89]),
            standard_name="surface_temperature",
            long_name=None,
            var_name="temp",
            units="K",
        )
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, least_significant_digit=1)
            cube_saved = iris.load_cube(nc_path)
            self.assertEqual(
                cube_saved.attributes["least_significant_digit"], 1
            )
            self.assertFalse(np.all(cube.data == cube_saved.data))
            self.assertArrayAllClose(cube.data, cube_saved.data, 0.1)

    def test_default_unlimited_dimensions(self):
        # Default is no unlimited dimensions.
        cube = self._simple_cube(">f4")
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            ds = nc.Dataset(nc_path)
            self.assertFalse(ds.dimensions["dim0"].isunlimited())
            self.assertFalse(ds.dimensions["dim1"].isunlimited())
            ds.close()

    def test_no_unlimited_dimensions(self):
        cube = self._simple_cube(">f4")
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=None)
            ds = nc.Dataset(nc_path)
            for dim in ds.dimensions.values():
                self.assertFalse(dim.isunlimited())
            ds.close()

    def test_invalid_unlimited_dimensions(self):
        cube = self._simple_cube(">f4")
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                # should not raise an exception
                saver.write(cube, unlimited_dimensions=["not_found"])

    def test_custom_unlimited_dimensions(self):
        cube = self._transverse_mercator_cube()
        unlimited_dimensions = [
            "projection_y_coordinate",
            "projection_x_coordinate",
        ]
        # test coordinates by name
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=unlimited_dimensions)
            ds = nc.Dataset(nc_path)
            for dim in unlimited_dimensions:
                self.assertTrue(ds.dimensions[dim].isunlimited())
            ds.close()
        # test coordinate arguments
        with self.temp_filename(".nc") as nc_path:
            coords = [cube.coord(dim) for dim in unlimited_dimensions]
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=coords)
            ds = nc.Dataset(nc_path)
            for dim in unlimited_dimensions:
                self.assertTrue(ds.dimensions[dim].isunlimited())
            ds.close()

    def test_reserved_attributes(self):
        cube = self._simple_cube(">f4")
        cube.attributes["dimensions"] = "something something_else"
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            ds = nc.Dataset(nc_path)
            res = ds.getncattr("dimensions")
            ds.close()
            self.assertEqual(res, "something something_else")

    def test_with_climatology(self):
        cube = stock.climatology_3d()
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube)
            self.assertCDL(nc_path)


class Test__create_cf_bounds(tests.IrisTest):
    def _check_bounds_setting(self, climatological=False):
        # Generic test that can run with or without a climatological coord.
        cube = stock.climatology_3d()
        coord = cube.coord("time").copy()
        # Over-write original value from stock.climatology_3d with test value.
        coord.climatological = climatological

        # Set up expected strings.
        if climatological:
            property_name = "climatology"
            varname_extra = "climatology"
        else:
            property_name = "bounds"
            varname_extra = "bnds"
        boundsvar_name = "time_" + varname_extra

        # Set up arguments for testing _create_cf_bounds.
        saver = mock.MagicMock(spec=Saver)
        # NOTE: 'saver' must have spec=Saver to fake isinstance(save, Saver),
        # so it can pass as 'self' in the call to _create_cf_cbounds.
        # Mock a '_dataset' property; not automatic because 'spec=Saver'.
        saver._dataset = mock.MagicMock()
        # Mock the '_ensure_valid_dtype' method to return an object with a
        # suitable 'shape' and 'dtype'.
        saver._ensure_valid_dtype.return_value = mock.Mock(
            shape=coord.bounds.shape, dtype=coord.bounds.dtype
        )
        var = mock.MagicMock(spec=nc.Variable)

        # Make the main call.
        Saver._create_cf_bounds(saver, coord, var, "time")

        # Test the call of _setncattr in _create_cf_bounds.
        setncattr_call = mock.call(
            property_name, boundsvar_name.encode(encoding="ascii")
        )
        self.assertEqual(setncattr_call, var.setncattr.call_args)

        # Test the call of createVariable in _create_cf_bounds.
        dataset = saver._dataset
        expected_dimensions = var.dimensions + ("bnds",)
        create_var_call = mock.call(
            boundsvar_name, coord.bounds.dtype, expected_dimensions
        )
        self.assertEqual(create_var_call, dataset.createVariable.call_args)

    def test_set_bounds_default(self):
        self._check_bounds_setting(climatological=False)

    def test_set_bounds_climatology(self):
        self._check_bounds_setting(climatological=True)


class Test_write__valid_x_cube_attributes(tests.IrisTest):
    """Testing valid_range, valid_min and valid_max attributes."""

    def test_valid_range_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        vrange = np.array([1, 2], dtype="int32")
        cube.attributes["valid_range"] = vrange
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(ds.valid_range, vrange)
            ds.close()

    def test_valid_min_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        cube.attributes["valid_min"] = 1
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(ds.valid_min, 1)
            ds.close()

    def test_valid_max_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        cube.attributes["valid_max"] = 2
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(ds.valid_max, 2)
            ds.close()


class Test_write__valid_x_coord_attributes(tests.IrisTest):
    """Testing valid_range, valid_min and valid_max attributes."""

    def test_valid_range_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        vrange = np.array([1, 2], dtype="int32")
        cube.coord(axis="x").attributes["valid_range"] = vrange
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(
                ds.variables["longitude"].valid_range, vrange
            )
            ds.close()

    def test_valid_min_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        cube.coord(axis="x").attributes["valid_min"] = 1
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(ds.variables["longitude"].valid_min, 1)
            ds.close()

    def test_valid_max_saved(self):
        cube = tests.stock.lat_lon_cube()
        cube.data = cube.data.astype("int32")

        cube.coord(axis="x").attributes["valid_max"] = 2
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, unlimited_dimensions=[])
            ds = nc.Dataset(nc_path)
            self.assertArrayEqual(ds.variables["longitude"].valid_max, 2)
            ds.close()


class Test_write_fill_value(tests.IrisTest):
    def _make_cube(
        self, dtype, lazy=False, masked_value=None, masked_index=None
    ):
        data = np.arange(12, dtype=dtype).reshape(3, 4)
        if masked_value is not None:
            data = ma.masked_equal(data, masked_value)
        if masked_index is not None:
            data = np.ma.masked_array(data)
            data[masked_index] = ma.masked
        if lazy:
            data = as_lazy_data(data)
        lat = DimCoord(np.arange(3), "latitude", units="degrees")
        lon = DimCoord(np.arange(4), "longitude", units="degrees")
        return Cube(
            data,
            standard_name="air_temperature",
            units="K",
            dim_coords_and_dims=[(lat, 0), (lon, 1)],
        )

    @contextmanager
    def _netCDF_var(self, cube, **kwargs):
        # Get the netCDF4 Variable for a cube from a temp file
        standard_name = cube.standard_name
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(cube, **kwargs)
            ds = nc.Dataset(nc_path)
            (var,) = [
                var
                for var in ds.variables.values()
                if var.standard_name == standard_name
            ]
            yield var

    def test_fill_value(self):
        # Test that a passed fill value is saved as a _FillValue attribute.
        cube = self._make_cube(">f4")
        fill_value = 12345.0
        with self._netCDF_var(cube, fill_value=fill_value) as var:
            self.assertEqual(fill_value, var._FillValue)

    def test_default_fill_value(self):
        # Test that if no fill value is passed then there is no _FillValue.
        # attribute.
        cube = self._make_cube(">f4")
        with self._netCDF_var(cube) as var:
            self.assertNotIn("_FillValue", var.ncattrs())

    def test_mask_fill_value(self):
        # Test that masked data saves correctly when given a fill value.
        index = (1, 1)
        fill_value = 12345.0
        cube = self._make_cube(">f4", masked_index=index)
        with self._netCDF_var(cube, fill_value=fill_value) as var:
            self.assertEqual(fill_value, var._FillValue)
            self.assertTrue(var[index].mask)

    def test_mask_default_fill_value(self):
        # Test that masked data saves correctly using the default fill value.
        index = (1, 1)
        cube = self._make_cube(">f4", masked_index=index)
        with self._netCDF_var(cube) as var:
            self.assertNotIn("_FillValue", var.ncattrs())
            self.assertTrue(var[index].mask)

    def test_mask_lazy_fill_value(self):
        # Test that masked lazy data saves correctly when given a fill value.
        index = (1, 1)
        fill_value = 12345.0
        cube = self._make_cube(">f4", masked_index=index, lazy=True)
        with self._netCDF_var(cube, fill_value=fill_value) as var:
            self.assertEqual(var._FillValue, fill_value)
            self.assertTrue(var[index].mask)

    def test_mask_lazy_default_fill_value(self):
        # Test that masked lazy data saves correctly using the default fill
        # value.
        index = (1, 1)
        cube = self._make_cube(">f4", masked_index=index, lazy=True)
        with self._netCDF_var(cube) as var:
            self.assertNotIn("_FillValue", var.ncattrs())
            self.assertTrue(var[index].mask)

    def test_contains_fill_value_passed(self):
        # Test that a warning is raised if the data contains the fill value.
        cube = self._make_cube(">f4")
        fill_value = 1
        with self.assertWarnsRegexp(
            "contains unmasked data points equal to the fill-value"
        ):
            with self._netCDF_var(cube, fill_value=fill_value):
                pass

    def test_contains_fill_value_byte(self):
        # Test that a warning is raised if the data contains the fill value
        # when it is of a byte type.
        cube = self._make_cube(">i1")
        fill_value = 1
        with self.assertWarnsRegexp(
            "contains unmasked data points equal to the fill-value"
        ):
            with self._netCDF_var(cube, fill_value=fill_value):
                pass

    def test_contains_default_fill_value(self):
        # Test that a warning is raised if the data contains the default fill
        # value if no fill_value argument is supplied.
        cube = self._make_cube(">f4")
        cube.data[0, 0] = nc.default_fillvals["f4"]
        with self.assertWarnsRegexp(
            "contains unmasked data points equal to the fill-value"
        ):
            with self._netCDF_var(cube):
                pass

    def test_contains_default_fill_value_byte(self):
        # Test that no warning is raised if the data contains the default fill
        # value if no fill_value argument is supplied when the data is of a
        # byte type.
        cube = self._make_cube(">i1")
        with self.assertNoWarningsRegexp(r"\(fill\|mask\)"):
            with self._netCDF_var(cube):
                pass

    def test_contains_masked_fill_value(self):
        # Test that no warning is raised if the data contains the fill_value at
        # a masked point.
        fill_value = 1
        cube = self._make_cube(">f4", masked_value=fill_value)
        with self.assertNoWarningsRegexp(r"\(fill\|mask\)"):
            with self._netCDF_var(cube, fill_value=fill_value):
                pass

    def test_masked_byte_default_fill_value(self):
        # Test that a warning is raised when saving masked byte data with no
        # fill value supplied.
        cube = self._make_cube(">i1", masked_value=1)
        with self.assertNoWarningsRegexp(r"\(fill\|mask\)"):
            with self._netCDF_var(cube):
                pass

    def test_masked_byte_fill_value_passed(self):
        # Test that no warning is raised when saving masked byte data with a
        # fill value supplied if the the data does not contain the fill_value.
        fill_value = 100
        cube = self._make_cube(">i1", masked_value=2)
        with self.assertNoWarningsRegexp(r"\(fill\|mask\)"):
            with self._netCDF_var(cube, fill_value=fill_value):
                pass


class Test_cf_valid_var_name(tests.IrisTest):
    def test_no_replacement(self):
        self.assertEqual(Saver.cf_valid_var_name("valid_Nam3"), "valid_Nam3")

    def test_special_chars(self):
        self.assertEqual(Saver.cf_valid_var_name("inv?alid"), "inv_alid")

    def test_leading_underscore(self):
        self.assertEqual(Saver.cf_valid_var_name("_invalid"), "var__invalid")

    def test_leading_number(self):
        self.assertEqual(Saver.cf_valid_var_name("2invalid"), "var_2invalid")

    def test_leading_invalid(self):
        self.assertEqual(Saver.cf_valid_var_name("?invalid"), "var__invalid")

    def test_no_hyphen(self):
        # CF explicitly prohibits hyphen, even though it is fine in NetCDF.
        self.assertEqual(
            Saver.cf_valid_var_name("valid-netcdf"), "valid_netcdf"
        )


class _Common__check_attribute_compliance:
    def setUp(self):
        self.container = mock.Mock(name="container", attributes={})
        self.data = np.array(1, dtype="int32")

        patch = mock.patch("netCDF4.Dataset")
        _ = patch.start()
        self.addCleanup(patch.stop)

    def set_attribute(self, value):
        self.container.attributes[self.attribute] = value

    def assertAttribute(self, value):
        self.assertEqual(
            np.asarray(self.container.attributes[self.attribute]).dtype, value
        )

    def check_attribute_compliance_call(self, value):
        self.set_attribute(value)
        with Saver(mock.Mock(), "NETCDF4") as saver:
            saver.check_attribute_compliance(self.container, self.data)


class Test_check_attribute_compliance__valid_range(
    _Common__check_attribute_compliance, tests.IrisTest
):
    @property
    def attribute(self):
        return "valid_range"

    def test_valid_range_type_coerce(self):
        value = np.array([1, 2], dtype="float")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(self.data.dtype)

    def test_valid_range_unsigned_int8_data_signed_range(self):
        self.data = self.data.astype("uint8")
        value = np.array([1, 2], dtype="int8")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(value.dtype)

    def test_valid_range_cannot_coerce(self):
        value = np.array([1.5, 2.5], dtype="float64")
        msg = '"valid_range" is not of a suitable value'
        with self.assertRaisesRegex(ValueError, msg):
            self.check_attribute_compliance_call(value)

    def test_valid_range_not_numpy_array(self):
        # Ensure we handle the case when not a numpy array is provided.
        self.data = self.data.astype("int8")
        value = [1, 2]
        self.check_attribute_compliance_call(value)
        self.assertAttribute(np.int64)


class Test_check_attribute_compliance__valid_min(
    _Common__check_attribute_compliance, tests.IrisTest
):
    @property
    def attribute(self):
        return "valid_min"

    def test_valid_range_type_coerce(self):
        value = np.array(1, dtype="float")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(self.data.dtype)

    def test_valid_range_unsigned_int8_data_signed_range(self):
        self.data = self.data.astype("uint8")
        value = np.array(1, dtype="int8")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(value.dtype)

    def test_valid_range_cannot_coerce(self):
        value = np.array(1.5, dtype="float64")
        msg = '"valid_min" is not of a suitable value'
        with self.assertRaisesRegex(ValueError, msg):
            self.check_attribute_compliance_call(value)

    def test_valid_range_not_numpy_array(self):
        # Ensure we handle the case when not a numpy array is provided.
        self.data = self.data.astype("int8")
        value = 1
        self.check_attribute_compliance_call(value)
        self.assertAttribute(np.int64)


class Test_check_attribute_compliance__valid_max(
    _Common__check_attribute_compliance, tests.IrisTest
):
    @property
    def attribute(self):
        return "valid_max"

    def test_valid_range_type_coerce(self):
        value = np.array(2, dtype="float")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(self.data.dtype)

    def test_valid_range_unsigned_int8_data_signed_range(self):
        self.data = self.data.astype("uint8")
        value = np.array(2, dtype="int8")
        self.check_attribute_compliance_call(value)
        self.assertAttribute(value.dtype)

    def test_valid_range_cannot_coerce(self):
        value = np.array(2.5, dtype="float64")
        msg = '"valid_max" is not of a suitable value'
        with self.assertRaisesRegex(ValueError, msg):
            self.check_attribute_compliance_call(value)

    def test_valid_range_not_numpy_array(self):
        # Ensure we handle the case when not a numpy array is provided.
        self.data = self.data.astype("int8")
        value = 2
        self.check_attribute_compliance_call(value)
        self.assertAttribute(np.int64)


class Test_check_attribute_compliance__exception_handlng(
    _Common__check_attribute_compliance, tests.IrisTest
):
    def test_valid_range_and_valid_min_valid_max_provided(self):
        # Conflicting attributes should raise a suitable exception.
        self.data = self.data.astype("int8")
        self.container.attributes["valid_range"] = [1, 2]
        self.container.attributes["valid_min"] = [1]
        msg = 'Both "valid_range" and "valid_min"'
        with Saver(mock.Mock(), "NETCDF4") as saver:
            with self.assertRaisesRegex(ValueError, msg):
                saver.check_attribute_compliance(self.container, self.data)


class Test__cf_coord_identity(tests.IrisTest):
    def check_call(self, coord_name, coord_system, units, expected_units):
        coord = iris.coords.DimCoord(
            [30, 45], coord_name, units=units, coord_system=coord_system
        )
        result = Saver._cf_coord_identity(coord)
        self.assertEqual(
            result, (coord.standard_name, coord.long_name, expected_units)
        )

    def test_geogcs_latitude(self):
        crs = iris.coord_systems.GeogCS(60, 0)
        self.check_call(
            "latitude",
            coord_system=crs,
            units="degrees",
            expected_units="degrees_north",
        )

    def test_geogcs_longitude(self):
        crs = iris.coord_systems.GeogCS(60, 0)
        self.check_call(
            "longitude",
            coord_system=crs,
            units="degrees",
            expected_units="degrees_east",
        )

    def test_no_coord_system_latitude(self):
        self.check_call(
            "latitude",
            coord_system=None,
            units="degrees",
            expected_units="degrees_north",
        )

    def test_no_coord_system_longitude(self):
        self.check_call(
            "longitude",
            coord_system=None,
            units="degrees",
            expected_units="degrees_east",
        )

    def test_passthrough_units(self):
        crs = iris.coord_systems.LambertConformal(0, 20)
        self.check_call(
            "projection_x_coordinate",
            coord_system=crs,
            units="km",
            expected_units="km",
        )


class Test__create_cf_grid_mapping(tests.IrisTest):
    def _cube_with_cs(self, coord_system):
        """Return a simple 2D cube that uses the given coordinate system."""
        cube = stock.lat_lon_cube()
        x, y = cube.coord("longitude"), cube.coord("latitude")
        x.coord_system = y.coord_system = coord_system
        return cube

    def _grid_mapping_variable(self, coord_system):
        """
        Return a mock netCDF variable that represents the conversion
        of the given coordinate system.

        """
        cube = self._cube_with_cs(coord_system)

        class NCMock(mock.Mock):
            def setncattr(self, name, attr):
                setattr(self, name, attr)

        # Calls the actual NetCDF saver with appropriate mocking, returning
        # the grid variable that gets created.
        grid_variable = NCMock(name="NetCDFVariable")
        create_var_fn = mock.Mock(side_effect=[grid_variable])
        dataset = mock.Mock(variables=[], createVariable=create_var_fn)
        saver = mock.Mock(spec=Saver, _coord_systems=[], _dataset=dataset)
        variable = NCMock()

        # This is the method we're actually testing!
        Saver._create_cf_grid_mapping(saver, cube, variable)

        self.assertEqual(create_var_fn.call_count, 1)
        self.assertEqual(
            variable.grid_mapping, grid_variable.grid_mapping_name
        )
        return grid_variable

    def _variable_attributes(self, coord_system):
        """
        Return the attributes dictionary for the grid mapping variable
        that is created from the given coordinate system.

        """
        mock_grid_variable = self._grid_mapping_variable(coord_system)

        # Get the attributes defined on the mock object.
        attributes = sorted(mock_grid_variable.__dict__.keys())
        attributes = [name for name in attributes if not name.startswith("_")]
        attributes.remove("method_calls")
        return {key: getattr(mock_grid_variable, key) for key in attributes}

    def _test(self, coord_system, expected):
        actual = self._variable_attributes(coord_system)

        # To see obvious differences, check that they keys are the same.
        self.assertEqual(sorted(actual.keys()), sorted(expected.keys()))
        # Now check that the values are equivalent.
        self.assertEqual(actual, expected)

    def test_rotated_geog_cs(self):
        coord_system = RotatedGeogCS(37.5, 177.5, ellipsoid=GeogCS(6371229.0))
        expected = {
            "grid_mapping_name": b"rotated_latitude_longitude",
            "north_pole_grid_longitude": 0.0,
            "grid_north_pole_longitude": 177.5,
            "grid_north_pole_latitude": 37.5,
            "longitude_of_prime_meridian": 0.0,
            "earth_radius": 6371229.0,
        }
        self._test(coord_system, expected)

    def test_spherical_geog_cs(self):
        coord_system = GeogCS(6371229.0)
        expected = {
            "grid_mapping_name": b"latitude_longitude",
            "longitude_of_prime_meridian": 0.0,
            "earth_radius": 6371229.0,
        }
        self._test(coord_system, expected)

    def test_elliptic_geog_cs(self):
        coord_system = GeogCS(637, 600)
        expected = {
            "grid_mapping_name": b"latitude_longitude",
            "longitude_of_prime_meridian": 0.0,
            "semi_minor_axis": 600.0,
            "semi_major_axis": 637.0,
        }
        self._test(coord_system, expected)

    def test_lambert_conformal(self):
        coord_system = LambertConformal(
            central_lat=44,
            central_lon=2,
            false_easting=-2,
            false_northing=-5,
            secant_latitudes=(38, 50),
            ellipsoid=GeogCS(6371000),
        )
        expected = {
            "grid_mapping_name": b"lambert_conformal_conic",
            "latitude_of_projection_origin": 44,
            "longitude_of_central_meridian": 2,
            "false_easting": -2,
            "false_northing": -5,
            "standard_parallel": (38, 50),
            "earth_radius": 6371000,
            "longitude_of_prime_meridian": 0,
        }
        self._test(coord_system, expected)

    def test_laea_cs(self):
        coord_system = LambertAzimuthalEqualArea(
            latitude_of_projection_origin=52,
            longitude_of_projection_origin=10,
            false_easting=100,
            false_northing=200,
            ellipsoid=GeogCS(6377563.396, 6356256.909),
        )
        expected = {
            "grid_mapping_name": b"lambert_azimuthal_equal_area",
            "latitude_of_projection_origin": 52,
            "longitude_of_projection_origin": 10,
            "false_easting": 100,
            "false_northing": 200,
            "semi_major_axis": 6377563.396,
            "semi_minor_axis": 6356256.909,
            "longitude_of_prime_meridian": 0,
        }
        self._test(coord_system, expected)

    def test_aea_cs(self):
        coord_system = AlbersEqualArea(
            latitude_of_projection_origin=52,
            longitude_of_central_meridian=10,
            false_easting=100,
            false_northing=200,
            standard_parallels=(38, 50),
            ellipsoid=GeogCS(6377563.396, 6356256.909),
        )
        expected = {
            "grid_mapping_name": b"albers_conical_equal_area",
            "latitude_of_projection_origin": 52,
            "longitude_of_central_meridian": 10,
            "false_easting": 100,
            "false_northing": 200,
            "standard_parallel": (38, 50),
            "semi_major_axis": 6377563.396,
            "semi_minor_axis": 6356256.909,
            "longitude_of_prime_meridian": 0,
        }
        self._test(coord_system, expected)

    def test_vp_cs(self):
        latitude_of_projection_origin = 1.0
        longitude_of_projection_origin = 2.0
        perspective_point_height = 2000000.0
        false_easting = 100.0
        false_northing = 200.0

        semi_major_axis = 6377563.396
        semi_minor_axis = 6356256.909
        ellipsoid = GeogCS(semi_major_axis, semi_minor_axis)

        coord_system = VerticalPerspective(
            latitude_of_projection_origin=latitude_of_projection_origin,
            longitude_of_projection_origin=longitude_of_projection_origin,
            perspective_point_height=perspective_point_height,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=ellipsoid,
        )
        expected = {
            "grid_mapping_name": b"vertical_perspective",
            "latitude_of_projection_origin": latitude_of_projection_origin,
            "longitude_of_projection_origin": longitude_of_projection_origin,
            "perspective_point_height": perspective_point_height,
            "false_easting": false_easting,
            "false_northing": false_northing,
            "semi_major_axis": semi_major_axis,
            "semi_minor_axis": semi_minor_axis,
            "longitude_of_prime_meridian": 0,
        }
        self._test(coord_system, expected)

    def test_geo_cs(self):
        latitude_of_projection_origin = 0.0
        longitude_of_projection_origin = 2.0
        perspective_point_height = 2000000.0
        sweep_angle_axis = "x"
        false_easting = 100.0
        false_northing = 200.0

        semi_major_axis = 6377563.396
        semi_minor_axis = 6356256.909
        ellipsoid = GeogCS(semi_major_axis, semi_minor_axis)

        coord_system = Geostationary(
            latitude_of_projection_origin=latitude_of_projection_origin,
            longitude_of_projection_origin=longitude_of_projection_origin,
            perspective_point_height=perspective_point_height,
            sweep_angle_axis=sweep_angle_axis,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=ellipsoid,
        )
        expected = {
            "grid_mapping_name": b"geostationary",
            "latitude_of_projection_origin": latitude_of_projection_origin,
            "longitude_of_projection_origin": longitude_of_projection_origin,
            "perspective_point_height": perspective_point_height,
            "sweep_angle_axis": sweep_angle_axis,
            "false_easting": false_easting,
            "false_northing": false_northing,
            "semi_major_axis": semi_major_axis,
            "semi_minor_axis": semi_minor_axis,
            "longitude_of_prime_meridian": 0,
        }
        self._test(coord_system, expected)


class Test__create_cf_cell_measure_variable(tests.IrisTest):
    # Saving of masked data is disallowed.
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        self.names_map = ["latitude", "longitude"]
        masked_array = np.ma.masked_array([0, 1, 2], mask=[True, False, True])
        self.cm = iris.coords.CellMeasure(masked_array, var_name="cell_area")
        self.cube.add_cell_measure(self.cm, data_dims=0)
        self.exp_emsg = "Cell measures with missing data are not supported."

    def test_masked_data__insitu(self):
        # Test that the error is raised in the right place.
        with self.temp_filename(".nc") as nc_path:
            saver = Saver(nc_path, "NETCDF4")
            with self.assertRaisesRegex(ValueError, self.exp_emsg):
                saver._create_cf_cell_measure_variable(
                    self.cube, self.names_map, self.cm
                )

    def test_masked_data__save_pipeline(self):
        # Test that the right error is raised by the saver pipeline.
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                with self.assertRaisesRegex(ValueError, self.exp_emsg):
                    saver.write(self.cube)


if __name__ == "__main__":
    tests.main()



"""
Seasonal Ensemble Model Plots
=============================

This example demonstrates the loading of a lagged ensemble dataset from the
GloSea4 model, which is then used to produce two types of plot:

 * The first shows the "postage stamp" style image with an array of 14 images,
   one for each ensemble member with a shared colorbar. (The missing image in
   this example represents ensemble member number 6 which was a failed run)

 * The second plot shows the data limited to a region of interest, in this case
   a region defined for forecasting ENSO (El Nino-Southern Oscillation), which,
   for the purposes of this example, has had the ensemble mean subtracted from
   each ensemble member to give an anomaly surface temperature. In practice a
   better approach would be to take the climatological mean, calibrated to the
   model, from each ensemble member.

"""

import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.plot as iplt


def realization_metadata(cube, field, fname):
    """
    A function which modifies the cube's metadata to add a "realization"
    (ensemble member) coordinate from the filename if one doesn't already exist
    in the cube.

    """
    # add an ensemble member coordinate if one doesn't already exist
    if not cube.coords("realization"):
        # the ensemble member is encoded in the filename as *_???.pp where ???
        # is the ensemble member
        realization_number = fname[-6:-3]

        import iris.coords

        realization_coord = iris.coords.AuxCoord(
            np.int32(realization_number), "realization", units="1"
        )
        cube.add_aux_coord(realization_coord)


def main():
    # extract surface temperature cubes which have an ensemble member
    # coordinate, adding appropriate lagged ensemble metadata
    surface_temp = iris.load_cube(
        iris.sample_data_path("GloSea4", "ensemble_???.pp"),
        iris.Constraint("surface_temperature", realization=lambda value: True),
        callback=realization_metadata,
    )

    # -------------------------------------------------------------------------
    # Plot #1: Ensemble postage stamps
    # -------------------------------------------------------------------------

    # for the purposes of this example, take the last time element of the cube
    last_timestep = surface_temp[:, -1, :, :]

    # Make 50 evenly spaced levels which span the dataset
    contour_levels = np.linspace(
        np.min(last_timestep.data), np.max(last_timestep.data), 50
    )

    # Create a wider than normal figure to support our many plots
    plt.figure(figsize=(12, 6), dpi=100)

    # Also manually adjust the spacings which are used when creating subplots
    plt.gcf().subplots_adjust(
        hspace=0.05,
        wspace=0.05,
        top=0.95,
        bottom=0.05,
        left=0.075,
        right=0.925,
    )

    # iterate over all possible latitude longitude slices
    for cube in last_timestep.slices(["latitude", "longitude"]):

        # get the ensemble member number from the ensemble coordinate
        ens_member = cube.coord("realization").points[0]

        # plot the data in a 4x4 grid, with each plot's position in the grid
        # being determined by ensemble member number the special case for the
        # 13th ensemble member is to have the plot at the bottom right
        if ens_member == 13:
            plt.subplot(4, 4, 16)
        else:
            plt.subplot(4, 4, ens_member + 1)

        cf = iplt.contourf(cube, contour_levels)

        # add coastlines
        plt.gca().coastlines()

    # make an axes to put the shared colorbar in
    colorbar_axes = plt.gcf().add_axes([0.35, 0.1, 0.3, 0.05])
    colorbar = plt.colorbar(cf, colorbar_axes, orientation="horizontal")
    colorbar.set_label("%s" % last_timestep.units)

    # limit the colorbar to 8 tick marks
    import matplotlib.ticker

    colorbar.locator = matplotlib.ticker.MaxNLocator(8)
    colorbar.update_ticks()

    # get the time for the entire plot
    time_coord = last_timestep.coord("time")
    time = time_coord.units.num2date(time_coord.bounds[0, 0])

    # set a global title for the postage stamps with the date formated by
    # "monthname year"
    plt.suptitle(
        "Surface temperature ensemble forecasts for %s"
        % (time.strftime("%B %Y"),)
    )

    iplt.show()

    # -------------------------------------------------------------------------
    # Plot #2: ENSO plumes
    # -------------------------------------------------------------------------

    # Nino 3.4 lies between: 170W and 120W, 5N and 5S, so define a constraint
    # which matches this
    nino_3_4_constraint = iris.Constraint(
        longitude=lambda v: -170 + 360 <= v <= -120 + 360,
        latitude=lambda v: -5 <= v <= 5,
    )

    nino_cube = surface_temp.extract(nino_3_4_constraint)

    # Subsetting a circular longitude coordinate always results in a circular
    # coordinate, so set the coordinate to be non-circular
    nino_cube.coord("longitude").circular = False

    # Calculate the horizontal mean for the nino region
    mean = nino_cube.collapsed(["latitude", "longitude"], iris.analysis.MEAN)

    # Calculate the ensemble mean of the horizontal mean. To do this, remove
    # the "forecast_period" and "forecast_reference_time" coordinates which
    # span both "relalization" and "time".
    mean.remove_coord("forecast_reference_time")
    mean.remove_coord("forecast_period")
    ensemble_mean = mean.collapsed("realization", iris.analysis.MEAN)

    # take the ensemble mean from each ensemble member
    mean -= ensemble_mean.data

    plt.figure()

    for ensemble_member in mean.slices(["time"]):
        # draw each ensemble member as a dashed line in black
        iplt.plot(ensemble_member, "--k")

    plt.title("Mean temperature anomaly for ENSO 3.4 region")
    plt.xlabel("Time")
    plt.ylabel("Temperature anomaly / K")

    iplt.show()


if __name__ == "__main__":
    main()



"""
Applying a Filter to a Time-Series
==================================

This example demonstrates low pass filtering a time-series by applying a
weighted running mean over the time dimension.

The time-series used is the Darwin-only Southern Oscillation index (SOI),
which is filtered using two different Lanczos filters, one to filter out
time-scales of less than two years and one to filter out time-scales of
less than 7 years.

References
----------

    Duchon C. E. (1979) Lanczos Filtering in One and Two Dimensions.
    Journal of Applied Meteorology, Vol 18, pp 1016-1022.

    Trenberth K. E. (1984) Signal Versus Noise in the Southern Oscillation.
    Monthly Weather Review, Vol 112, pp 326-332

"""

import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.plot as iplt


def low_pass_weights(window, cutoff):
    """Calculate weights for a low pass Lanczos filter.

    Args:

    window: int
        The length of the filter window.

    cutoff: float
        The cutoff frequency in inverse time steps.

    """
    order = ((window - 1) // 2) + 1
    nwts = 2 * order + 1
    w = np.zeros([nwts])
    n = nwts // 2
    w[n] = 2 * cutoff
    k = np.arange(1.0, n)
    sigma = np.sin(np.pi * k / n) * n / (np.pi * k)
    firstfactor = np.sin(2.0 * np.pi * cutoff * k) / (np.pi * k)
    w[n - 1 : 0 : -1] = firstfactor * sigma
    w[n + 1 : -1] = firstfactor * sigma
    return w[1:-1]


def main():
    # Load the monthly-valued Southern Oscillation Index (SOI) time-series.
    fname = iris.sample_data_path("SOI_Darwin.nc")
    soi = iris.load_cube(fname)

    # Window length for filters.
    window = 121

    # Construct 2-year (24-month) and 7-year (84-month) low pass filters
    # for the SOI data which is monthly.
    wgts24 = low_pass_weights(window, 1.0 / 24.0)
    wgts84 = low_pass_weights(window, 1.0 / 84.0)

    # Apply each filter using the rolling_window method used with the weights
    # keyword argument. A weighted sum is required because the magnitude of
    # the weights are just as important as their relative sizes.
    soi24 = soi.rolling_window(
        "time", iris.analysis.SUM, len(wgts24), weights=wgts24
    )
    soi84 = soi.rolling_window(
        "time", iris.analysis.SUM, len(wgts84), weights=wgts84
    )

    # Plot the SOI time series and both filtered versions.
    plt.figure(figsize=(9, 4))
    iplt.plot(
        soi,
        color="0.7",
        linewidth=1.0,
        linestyle="-",
        alpha=1.0,
        label="no filter",
    )
    iplt.plot(
        soi24,
        color="b",
        linewidth=2.0,
        linestyle="-",
        alpha=0.7,
        label="2-year filter",
    )
    iplt.plot(
        soi84,
        color="r",
        linewidth=2.0,
        linestyle="-",
        alpha=0.7,
        label="7-year filter",
    )
    plt.ylim([-4, 4])
    plt.title("Southern Oscillation Index (Darwin Only)")
    plt.xlabel("Time")
    plt.ylabel("SOI")
    plt.legend(fontsize=10)
    iplt.show()


if __name__ == "__main__":
    main()



"""
Multi-Line Temperature Profile Plot
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

"""

import matplotlib.pyplot as plt

import iris
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    fname = iris.sample_data_path("air_temp.pp")

    # Load exactly one cube from the given file.
    temperature = iris.load_cube(fname)

    # We only want a small number of latitudes, so filter some out
    # using "extract".
    temperature = temperature.extract(
        iris.Constraint(latitude=lambda cell: 68 <= cell < 78)
    )

    for cube in temperature.slices("longitude"):

        # Create a string label to identify this cube (i.e. latitude: value).
        cube_label = "latitude: %s" % cube.coord("latitude").points[0]

        # Plot the cube, and associate it with a label.
        qplt.plot(cube, label=cube_label)

    # Add the legend with 2 columns.
    plt.legend(ncol=2)

    # Put a grid on the plot.
    plt.grid(True)

    # Tell matplotlib not to extend the plot axes range to nicely
    # rounded numbers.
    plt.axis("tight")

    # Finally, show it.
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.PROPORTION` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy.ma as ma

from iris.analysis import PROPORTION
import iris.cube
from iris.coords import DimCoord


class Test_units_func(tests.IrisTest):
    def test(self):
        self.assertIsNotNone(PROPORTION.units_func)
        new_units = PROPORTION.units_func(None)
        self.assertEqual(new_units, 1)


class Test_masked(tests.IrisTest):
    def setUp(self):
        self.cube = iris.cube.Cube(ma.masked_equal([1, 2, 3, 4, 5], 3))
        self.cube.add_dim_coord(DimCoord([6, 7, 8, 9, 10], long_name="foo"), 0)
        self.func = lambda x: x >= 3

    def test_ma(self):
        cube = self.cube.collapsed("foo", PROPORTION, function=self.func)
        self.assertArrayEqual(cube.data, [0.5])

    def test_false_mask(self):
        # Test corner case where mask is returned as boolean value rather
        # than boolean array when the mask is unspecified on construction.
        masked_cube = iris.cube.Cube(ma.array([1, 2, 3, 4, 5]))
        masked_cube.add_dim_coord(
            DimCoord([6, 7, 8, 9, 10], long_name="foo"), 0
        )
        cube = masked_cube.collapsed("foo", PROPORTION, function=self.func)
        self.assertArrayEqual(cube.data, ma.array([0.6]))


class Test_name(tests.IrisTest):
    def test(self):
        self.assertEqual(PROPORTION.name(), "proportion")


class Test_aggregate_shape(tests.IrisTest):
    def test(self):
        shape = ()
        kwargs = dict()
        self.assertTupleEqual(PROPORTION.aggregate_shape(**kwargs), shape)
        kwargs = dict(captain="caveman", penelope="pitstop")
        self.assertTupleEqual(PROPORTION.aggregate_shape(**kwargs), shape)


if __name__ == "__main__":
    tests.main()



import iris
import iris.plot as iplt
import matplotlib.pyplot as plt

# Load the data.
global_air_temp = iris.load_cube(iris.sample_data_path("air_temp.pp"))
rotated_psl = iris.load_cube(iris.sample_data_path("rotated_pole.nc"))

plt.figure(figsize=(9, 3.5))

plt.subplot(1, 2, 1)
iplt.pcolormesh(global_air_temp, norm=plt.Normalize(260, 300))
plt.title("Air temperature\n" "on a global longitude latitude grid")
ax = plt.gca()
ax.coastlines()
ax.gridlines()

plt.subplot(1, 2, 2)
iplt.pcolormesh(rotated_psl)
plt.title("Air pressure\n" "on a limited area rotated pole grid")
ax = plt.gca()
ax.coastlines(resolution="50m")
ax.gridlines()

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.analysis.name_loaders._cf_height_from_name`
function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.coords import AuxCoord
from iris.fileformats.name_loaders import _cf_height_from_name


class TestAll(tests.IrisTest):
    def _default_coord(self, data):
        # This private method returns a coordinate with values expected
        # when no interpretation is made of the field header string.
        return AuxCoord(
            units="no-unit",
            points=data,
            bounds=None,
            standard_name=None,
            long_name="z",
        )


class TestAll_NAMEII(TestAll):
    # NAMEII formats are defined by bounds, not points
    def test_bounded_height_above_ground(self):
        data = "From     0 -   100m agl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name="height",
            long_name="height above ground level",
        )
        self.assertEqual(com, res)

    def test_bounded_flight_level(self):
        data = "From FL0 - FL100"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="unknown",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name=None,
            long_name="flight_level",
        )
        self.assertEqual(com, res)

    def test_bounded_height_above_sea_level(self):
        data = "From     0 -   100m asl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name="altitude",
            long_name="altitude above sea level",
        )
        self.assertEqual(com, res)

    def test_malformed_height_above_ground(self):
        # Parse height above ground level with additional stuff on the end of
        # the string (agl).
        data = "From     0 -   100m agl and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_malformed_height_above_sea_level(self):
        # Parse height above ground level with additional stuff on the end of
        # the string (agl).
        data = "From     0 -   100m asl and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_malformed_flight_level(self):
        # Parse height above ground level with additional stuff on the end of
        # the string (agl).
        data = "From FL0 - FL100 and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_float_bounded_height_above_ground(self):
        # Parse height above ground level when its a float.
        data = "From     0.0 -   100.0m agl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name="height",
            long_name="height above ground level",
        )
        self.assertEqual(com, res)

    def test_float_bounded_height_flight_level(self):
        # Parse height above ground level, as a float (agl).
        data = "From FL0.0 - FL100.0"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="unknown",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name=None,
            long_name="flight_level",
        )
        self.assertEqual(com, res)

    def test_float_bounded_height_above_sea_level(self):
        # Parse height above ground level as a float (agl).
        data = "From     0.0 -   100.0m asl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name="altitude",
            long_name="altitude above sea level",
        )
        self.assertEqual(com, res)

    def test_no_match(self):
        # Parse height information when there is no match.
        # No interpretation, just returns default values.
        data = "Vertical integral"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_pressure(self):
        # Parse air_pressure string.
        data = "From     0 -   100 Pa"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="Pa",
            points=50.0,
            bounds=np.array([0.0, 100.0]),
            standard_name="air_pressure",
            long_name=None,
        )
        self.assertEqual(com, res)


class TestAll_NAMEIII(TestAll):
    # NAMEIII formats are defined by points, not bounds.
    def test_height_above_ground(self):
        data = "Z = 50.00000 m agl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=None,
            standard_name="height",
            long_name="height above ground level",
        )
        self.assertEqual(com, res)

    def test_height_flight_level(self):
        data = "Z = 50.00000 FL"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="unknown",
            points=50.0,
            bounds=None,
            standard_name=None,
            long_name="flight_level",
        )
        self.assertEqual(com, res)

    def test_height_above_sea_level(self):
        data = "Z = 50.00000 m asl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=None,
            standard_name="altitude",
            long_name="altitude above sea level",
        )
        self.assertEqual(com, res)

    def test_malformed_height_above_ground(self):
        # Parse height above ground level, with additonal stuff at the string
        # end (agl).
        data = "Z = 50.00000 m agl and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_malformed_height_above_sea_level(self):
        # Parse height above ground level, with additional stuff at string
        # end (agl).
        data = "Z = 50.00000 m asl and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_malformed_flight_level(self):
        # Parse height above ground level (agl), with additional stuff at
        # string end.
        data = "Z = 50.00000 FL and stuff"
        res = _cf_height_from_name(data)
        com = self._default_coord(data)
        self.assertEqual(com, res)

    def test_integer_height_above_ground(self):
        # Parse height above ground level when its an integer.
        data = "Z = 50 m agl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=None,
            standard_name="height",
            long_name="height above ground level",
        )
        self.assertEqual(com, res)

    def test_integer_height_flight_level(self):
        # Parse flight level when its an integer.
        data = "Z = 50 FL"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="unknown",
            points=50.0,
            bounds=None,
            standard_name=None,
            long_name="flight_level",
        )
        self.assertEqual(com, res)

    def test_integer_height_above_sea_level(self):
        # Parse height above sea level (asl) when its an integer.
        data = "Z = 50 m asl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=50.0,
            bounds=None,
            standard_name="altitude",
            long_name="altitude above sea level",
        )
        self.assertEqual(com, res)

    def test_enotation_height_above_ground(self):
        # Parse height above ground expressed in scientific notation
        data = "Z = 0.0000000E+00 m agl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=0.0,
            bounds=None,
            standard_name="height",
            long_name="height above ground level",
        )
        self.assertEqual(com, res)

    def test_enotation_height_above_sea_level(self):
        # Parse height above sea level expressed in scientific notation
        data = "Z = 0.0000000E+00 m asl"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="m",
            points=0.0,
            bounds=None,
            standard_name="altitude",
            long_name="altitude above sea level",
        )
        self.assertEqual(com, res)

    def test_pressure(self):
        # Parse pressure.
        data = "Z = 50.00000 Pa"
        res = _cf_height_from_name(data)
        com = AuxCoord(
            units="Pa",
            points=50.0,
            bounds=None,
            standard_name="air_pressure",
            long_name=None,
        )
        self.assertEqual(com, res)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for subset."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.coords import DimCoord
from iris.cube import Cube


def _make_test_cube():
    data = np.zeros((4, 4, 1))
    lats, longs = [0, 10, 20, 30], [5, 15, 25, 35]
    lat_coord = DimCoord(lats, standard_name="latitude", units="degrees")
    lon_coord = DimCoord(longs, standard_name="longitude", units="degrees")
    vrt_coord = DimCoord([850], long_name="pressure", units="hPa")
    return Cube(
        data,
        long_name="test_cube",
        units="1",
        attributes=None,
        dim_coords_and_dims=[(lat_coord, 0), (lon_coord, 1)],
        aux_coords_and_dims=[(vrt_coord, None)],
    )


class TestSubset(tests.IrisTest):
    def setUp(self):
        self.cube = _make_test_cube()

    def test_coordinate_subset(self):
        coord = self.cube.coord("pressure")
        subsetted = self.cube.subset(coord)
        self.assertEqual(self.cube, subsetted)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_lambert_azimuthal_equal_area_coordinate_system`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import iris
from iris.coord_systems import LambertAzimuthalEqualArea
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_lambert_azimuthal_equal_area_coordinate_system


class TestBuildLambertAzimuthalEqualAreaCoordinateSystem(tests.IrisTest):
    def _test(self, inverse_flattening=False, no_optionals=False):
        if no_optionals:
            # Most properties are optional for this system.
            gridvar_props = {}
            # Setup all the expected default values
            test_lat = 0
            test_lon = 0
            test_easting = 0
            test_northing = 0
        else:
            # Choose test values and setup corresponding named properties.
            test_lat = -35
            test_lon = 175
            test_easting = -100
            test_northing = 200
            gridvar_props = dict(
                latitude_of_projection_origin=test_lat,
                longitude_of_projection_origin=test_lon,
                false_easting=test_easting,
                false_northing=test_northing)

        # Add ellipsoid args.
        gridvar_props['semi_major_axis'] = 6377563.396
        if inverse_flattening:
            gridvar_props['inverse_flattening'] = 299.3249646
            expected_ellipsoid = iris.coord_systems.GeogCS(
                6377563.396,
                inverse_flattening=299.3249646)
        else:
            gridvar_props['semi_minor_axis'] = 6356256.909
            expected_ellipsoid = iris.coord_systems.GeogCS(
                6377563.396, 6356256.909)

        cf_grid_var = mock.Mock(spec=[], **gridvar_props)

        cs = build_lambert_azimuthal_equal_area_coordinate_system(
            None, cf_grid_var)

        expected = LambertAzimuthalEqualArea(
            latitude_of_projection_origin=test_lat,
            longitude_of_projection_origin=test_lon,
            false_easting=test_easting,
            false_northing=test_northing,
            ellipsoid=expected_ellipsoid)

        self.assertEqual(cs, expected)

    def test_basic(self):
        self._test()

    def test_inverse_flattening(self):
        # Check when inverse_flattening is provided instead of semi_minor_axis.
        self._test(inverse_flattening=True)

    def test_no_optionals(self):
        # Check defaults, when all optional attributes are absent.
        self._test(no_optionals=True)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.describe_diff`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

from io import StringIO

import iris.cube
from iris.util import describe_diff


class Test(iris.tests.IrisTest):
    def setUp(self):
        self.cube_a = iris.cube.Cube([])
        self.cube_b = self.cube_a.copy()

    def _compare_result(self, cube_a, cube_b):
        result_sio = StringIO()
        describe_diff(cube_a, cube_b, output_file=result_sio)
        return result_sio.getvalue()

    def test_noncommon_array_attributes(self):
        # test non-common array attribute
        self.cube_a.attributes["test_array"] = np.array([1, 2, 3])
        return_str = self._compare_result(self.cube_a, self.cube_b)
        self.assertString(return_str, ["compatible_cubes.str.txt"])

    def test_same_array_attributes(self):
        # test matching array attribute
        self.cube_a.attributes["test_array"] = np.array([1, 2, 3])
        self.cube_b.attributes["test_array"] = np.array([1, 2, 3])
        return_str = self._compare_result(self.cube_a, self.cube_b)
        self.assertString(return_str, ["compatible_cubes.str.txt"])

    def test_different_array_attributes(self):
        # test non-matching array attribute
        self.cube_a.attributes["test_array"] = np.array([1, 2, 3])
        self.cube_b.attributes["test_array"] = np.array([1, 7, 3])
        return_str = self._compare_result(self.cube_a, self.cube_b)
        self.assertString(
            return_str,
            [
                "unit",
                "util",
                "describe_diff",
                "incompatible_array_attrs.str.txt",
            ],
        )


if __name__ == "__main__":
    tests.main()
